"""
Verifies that `.CpProcessPoolExecutor` handles all of its ``on_error`` modes
correctly.
"""

import os
import platform
import signal
import sys
import time
from concurrent.futures import CancelledError, TimeoutError
from concurrent.futures.process import EXTRA_QUEUED_CALLS
from multiprocessing import Manager

import pytest
from pyrseus.core.sys import SignalHandlerCtx
from pyrseus.ctx.mgr import ExecutorCtx

# These tests rely on SIGALRM. Unfortunately, that doesn't exist on Windows.
pytestmark = pytest.mark.skipif(
    platform.system() == "Windows", reason="SIGALRM does not exist on Windows"
)


def get_abbrev_state(fut):
    """
    Summarizes the state of a Future.
    """
    run = fut.running()
    done = fut.done()
    cancelled = False
    has_exc = False
    try:
        has_exc = fut.exception(0) is not None
    except TimeoutError:
        pass
    except CancelledError:
        cancelled = True
    return {
        # We only enumerate the combinations we actually expect to see in these
        # tests.
        (False, False, False, False): "pending",
        (True, False, False, False): "running",
        (False, True, False, False): "cancel_pending",
        (False, True, True, False): "cancelled",
        (False, False, True, False): "success",
        (True, False, False, True): "exception_pending",
        (False, False, True, True): "exception",
    }[run, cancelled, done, has_exc]


@pytest.mark.parametrize("on_error", ("wait", "cancel", "kill"))
def test_always_wait_if_no_error(on_error):
    """
    Tests that CpProcessPoolExecutor always waits till all submitted tasks enter
    a final state before exiting when there are no uncaught exceptions,
    regardless of the on_error mode.
    """

    # How many extra tasks concurrent.futures pre-queues to lower latency. These
    # can't be cancelled.
    assert 0 <= EXTRA_QUEUED_CALLS < 2

    # Called asynchronously 1s after all futures have been submitted.
    def release_barriers_on_sigalrm(signum, frame):
        assert signum == signal.SIGALRM
        # Verify that exactly 1 task is really running and is now waiting on our
        # barrier.
        for i, barrier in enumerate(barriers):
            if i == 0:
                assert barrier.n_waiting == 1
            else:
                assert barrier.n_waiting == 0
        # Meet each task at our barrier. We include a small scheduling timeout.
        for barrier in barriers:
            barrier.wait(0.1)

    # Submit some tasks that wait on a 1s alarm timer before being allowed to
    # proceed. While the timer is waiting, start exiting the context. Verify
    # that (a) ctx.__exit__ blocks until all futs are done, but (b) it's
    # otherwise running pretty quickly [otherwise we have to doubt our min
    # bound].
    with Manager() as sync_mgr:
        ctx = ExecutorCtx("cpprocess", 1, on_error=on_error)
        with SignalHandlerCtx(signal.SIGALRM, release_barriers_on_sigalrm):
            try:
                exe = ctx.__enter__()
                assert exe.submit(os.getpid).result() != os.getpid()  # warm up
                # Create one really-started task, fill the rest of the task
                # pre-queue, and create one really not started task.
                barriers = []
                futs = []
                for _ in range(1 + EXTRA_QUEUED_CALLS + 1):
                    barriers.append(sync_mgr.Barrier(2))
                    futs.append(exe.submit(barriers[-1].wait))
                t0 = time.time()
                signal.alarm(1)
            except:  # NOQA
                exc_info = sys.exc_info()
                ctx.__exit__(*exc_info)
                raise AssertionError("Exception not expected") from exc_info[1]
            else:
                ctx.__exit__(None, None, None)
            t1 = time.time()
            for fut in futs:
                state = get_abbrev_state(fut)
                if state == "exception":
                    exc = RuntimeError("Unexpected task exception")
                    raise exc from fut.exception()
                else:
                    assert state == "success"
            elapsed_secs = t1 - t0
            # Should block on the 1s alarm, but then finish all tasks quickly.
            assert 1.0 <= elapsed_secs <= 1.3


def test_wait_on_error():
    """
    Tests that CpProcessPoolExecutor with on_error="wait" actually waits on all
    submitted futures when there's an uncaught exception.
    """

    # This is similar to test_always_wait_if_no_error, except we raise an
    # exception right after we start releasing the barriers. Like it, this
    # should wait till all tasks are done. For clarity, the only other comments
    # in this test function are ones that highlight which lines differ from
    # test_always_wait_if_no_error's.

    assert 0 <= EXTRA_QUEUED_CALLS < 2

    def release_barriers_on_sigalrm(signum, frame):
        assert signum == signal.SIGALRM
        for i, barrier in enumerate(barriers):
            if i == 0:
                assert barrier.n_waiting == 1
            else:
                assert barrier.n_waiting == 0
        for barrier in barriers:
            barrier.wait(0.1)

    with Manager() as sync_mgr:
        ctx = ExecutorCtx("cpprocess", 1, on_error="wait")  # <--------- DIFFERS
        with SignalHandlerCtx(signal.SIGALRM, release_barriers_on_sigalrm):
            try:
                exe = ctx.__enter__()
                assert exe.submit(os.getpid).result() != os.getpid()
                barriers = []
                futs = []
                for _ in range(1 + EXTRA_QUEUED_CALLS + 1):
                    barriers.append(sync_mgr.Barrier(2))
                    futs.append(exe.submit(barriers[-1].wait))
                t0 = time.time()
                signal.alarm(1)  # <------------------------------------ DIFFERS
                exc = RuntimeError("start exiting early")  # <---------- DIFFERS
                raise exc  # <------------------------------------------ DIFFERS
            except:  # NOQA
                exc_info = sys.exc_info()
                ctx.__exit__(*exc_info)
                assert exc_info[1] is exc, exc_info[
                    1
                ]  # <---------------------------DIFFERS
            t1 = time.time()
            for fut in futs:
                state = get_abbrev_state(fut)
                if state == "exception":
                    exc = RuntimeError("Unexpected task exception")
                    raise exc from fut.exception()
                else:
                    assert state == "success"
            elapsed_secs = t1 - t0
            assert 1.0 <= elapsed_secs <= 1.3


def test_cancel_on_error(on_error="cancel"):
    """
    Tests that CpProcessPoolExecutor with on_error="cancel" waits on any running
    futures, but cancels all pending futures when there's an uncaught exception.
    """

    # This is similar to test_wait_on_error, except we use on_error="cancel".
    # We have highlighted the differences between that function and this one
    # with "DIFFERS" comments.

    assert 0 <= EXTRA_QUEUED_CALLS < 2

    def release_barriers_on_sigalrm(signum, frame):
        assert signum == signal.SIGALRM
        for i, barrier in enumerate(barriers):
            if i == 0:
                assert barrier.n_waiting == 1
            else:
                assert barrier.n_waiting == 0
        for barrier in barriers[: 1 + EXTRA_QUEUED_CALLS]:  # <--------- DIFFERS
            barrier.wait(0.1)

    with Manager() as sync_mgr:
        ctx = ExecutorCtx("cpprocess", 1, on_error=on_error)  # <------- DIFFERS
        with SignalHandlerCtx(signal.SIGALRM, release_barriers_on_sigalrm):
            try:
                exe = ctx.__enter__()
                assert exe.submit(os.getpid).result() != os.getpid()
                barriers = []
                futs = []
                for _ in range(1 + EXTRA_QUEUED_CALLS + 1):
                    barriers.append(sync_mgr.Barrier(2))
                    futs.append(exe.submit(barriers[-1].wait))
                t0 = time.time()
                signal.alarm(1)
                exc = RuntimeError("start exiting early")
                raise exc
            except:  # NOQA
                exc_info = sys.exc_info()
                ctx.__exit__(*exc_info)
                assert exc_info[1] is exc, exc_info[1]
            t1 = time.time()
            for fut in futs[: 1 + EXTRA_QUEUED_CALLS]:  # <------------- DIFFERS
                assert get_abbrev_state(fut) == "success"  # <---------- DIFFERS
            for fut in futs[1 + EXTRA_QUEUED_CALLS :]:  # <------------- DIFFERS
                assert get_abbrev_state(fut) == "cancelled"  # <-------- DIFFERS
            elapsed_secs = t1 - t0
            assert 1.0 <= elapsed_secs <= 1.2


def test_kill_on_error():
    """
    Tests that CpProcessPoolExecutor with on_error="kill" falls back to the
    on_error="cancel" behavior.
    """
    test_cancel_on_error(on_error="kill")



"""
A collection of helpers for writing `~pyrseus.ctx.api.ExecutorPluginEntryPoint`
subclasses that work with simple standard and semi-standard real-world
executors.
"""

import importlib
from concurrent.futures import ALL_COMPLETED, Executor, Future, wait
from functools import cached_property
from typing import Callable, Optional, Tuple, TypeVar

from pyrseus.core.sys import get_num_available_cores, module_exists
from pyrseus.ctx.api import (
    ExcInfo,
    ExecutorPluginEntryPoint,
    OnError,
    OnErrorLike,
    PreExitFunc,
    extract_keywords,
)

Ret = TypeVar("Ret")
"""
Represents the generic return type of a submitted callable.
"""


class OnExitProxy:
    """
    An `~concurrent.futures.Executor` proxy class that (a) tracks futures
    submitted to the specific executor, and (b) replaces the executor's
    ``__exit__`` method with a customized one from this class.

    - *When to Use:* This proxy is primarily useful for executors like |loky|_'s
      reusable ones. Without this wrapper, their context managers can
      unnecessarily wait on futures submitted by other managers or even
      submitted outside the context block. See the reusable variant of the
      `~pyrseus.ctx.plugins.loky` plugin for an example of using this proxy.

    - *When Not:* If you are using a normal executor where its context manager
      is the sole owner of its worker pool, considering using the lighter-weight
      `.shutdown_wait_on_error` and related functions instead of this proxy
      class.

    In the following example, ``e1_overblocks`` blocks until both its futures
    *and* ``e0``'s futures enter a final state:

        >>> from pyrseus.ctx.registry import skip_if_unavailable
        >>> import time

        >>> skip_if_unavailable("loky")
        >>> from loky import get_reusable_executor
        >>> t0 = time.time()
        >>> with get_reusable_executor() as e0:
        ...     _ = e0.submit(time.sleep, 1.0)
        ...     with get_reusable_executor() as e1_overblocks:
        ...         _ = e1_overblocks.submit(time.sleep, 0.1)
        ...     # e1_overblocks waited on e0's future too!
        ...     assert 1.0 <= time.time() - t0

    But with this wrapper, ``e1_precise`` only blocks on its own futures:

        >>> skip_if_unavailable("loky")
        >>> from loky import get_reusable_executor
        >>> with get_reusable_executor() as e0:
        ...     t0 = time.time()
        ...     _ = e0.submit(time.sleep, 1.0)
        ...     # Wrap the underlying executor in this proxy.
        ...     with OnExitProxy(
        ...         get_reusable_executor(),
        ...         on_error='cancel',
        ...         shutdown=False,
        ...     ) as e1_precise:
        ...         _ = e1_precise.submit(time.sleep, 0.1)
        ...     # e1_precise only blocked on its own future
        ...     assert 0.1 <= time.time() - t0 <= 0.9
        >>> # and e0 blocking still works as it should
        >>> assert 1.0 <= time.time() - t0

    :param exe: the executor to wrap. We forward all public attribute and method
        lookups to this object, not including special methods. This proxy is
        only designed for executors with trivial ``return self`` ``__enter__``
        methods.

    :param on_error: controls how ``exe`` and futures submitted via our
        `.submit` method are handled if an uncaught exception is passing through
        our `.__exit__` method. See `.OnError` for details. If ``shutdown`` is
        false, our `.__exit__` method uses `.wait` whenever possible instead of
        ``exe.shutdown(...)``.

    :param shutdown: whether to always call ``exe.shutdown(...)`` in our
        `.__exit__` method vs. try to use `.wait` instead when possible.
    """

    def __init__(self, exe: Executor, *, on_error: OnErrorLike, shutdown: bool):
        self._exe = exe
        self._on_error = OnError.get(on_error)
        self._shutdown = shutdown
        self._futs = set()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if self._shutdown:
            if exc_info == (None, None, None):
                self._exe.shutdown(wait=True)
            elif self._on_error is OnError.WAIT:
                self._exe.shutdown(wait=True)
            elif self._on_error is OnError.CANCEL_FUTURES:
                self._exe.shutdown(wait=True, cancel_futures=True)
            elif self._on_error is OnError.KILL_WORKERS:
                self._exe.shutdown(wait=True, kill_workers=True)
            else:
                raise ValueError(self._on_error)
        else:
            if exc_info == (None, None, None):
                wait(self._futs, return_when=ALL_COMPLETED)
            elif self._on_error is OnError.WAIT:
                wait(self._futs, return_when=ALL_COMPLETED)
            elif self._on_error is OnError.CANCEL_FUTURES:
                self.cancel_all()
                wait(self._futs, return_when=ALL_COMPLETED)
            elif self._on_error is OnError.KILL_WORKERS:
                # For now, we'll ask the executor to shut down everything, since
                # no major real-world executors publicly support killing just
                # the workers associated with specific tasks.
                self._exe.shutdown(wait=True, kill_workers=True)
            else:
                raise ValueError(self._on_error)

    def submit(self, fcn: Callable[..., Ret], /, *args, **kwargs) -> Future[Ret]:
        """
        Submits a task to the underlying executor, saving the future so that we
        can automatically block on it when the context is exited.
        """
        # Submit it to the underlying executor.
        fut = self._exe.submit(fcn, *args, **kwargs)
        # Save the future so that we can wait on it in __exit__ if needed.
        self._futs.add(fut)
        # Be nice and trim the set of tracked futures as fast as we can. This
        # way we don't leak memory if the results are large. This must come
        # after self._futs.add(fut) because the callback runs immediately if
        # already completed.
        fut.add_done_callback(lambda fut: self._futs.remove(fut))
        return fut

    def cancel_all(self) -> Tuple[int, int]:
        """
        Cancels all futures submitted through this executor proxy, but not any
        submitted to the underlying executor through other means, e.g. an
        aliased reference to the same |loky|_ reusable executor.

        :return: a tuple containing the number of futures that this method
            attempted to cancel, and the number it successfully cancelled.
            Futures that are already in a running or later state cannot be
            cancelled.
        """
        num_cancelled = 0
        futs = tuple(self._futs)  # make this race-free w.r.t. done callbacks
        for fut in futs:
            num_cancelled += fut.cancel()
        return len(futs), num_cancelled

    def __getattr__(self, name):
        """
        Makes this a perfect proxy class for public methods and attributes that
        aren't overridden in this proxy.
        """
        # For now, we'll only proxy public members, not special or non-public
        # ones. External users shouldn't be directly using non-public members.
        # Extra care has to be taken with some special members, so we don't even
        # try proxying those.
        if name.startswith("_"):
            if name not in dir(self._exe):
                raise AttributeError(name)
            else:
                raise AttributeError(f"Only public members are proxied, not: {name}")
        return getattr(self._exe, name)


def shutdown_wait_on_error(ctx: Executor, exc_info: ExcInfo):
    """
    Simple callback helper that invokes ``ctx.shutdown(wait=True)`` if
    ``exc_info`` is all ``None``.

    - *When to Use:* When making a plugin using `.SimpleEntryPoint`, consider
      having your `~.SimpleEntryPoint.wrap_for_wait_on_error` return this
      function as the second element of its returned tuple (the `.PreExitFunc`
      part). It will typically provide the desired behavior if ``ctx`` is the
      sole owner of its worker pool. See the `~pyrseus.ctx.plugins.process`
      plugin for an example.

    - *When Not:* For worker pools that may be shared between contexts, consider
      using `.OnExitProxy` to wrap the first returned element instead.
    """
    if exc_info != (None, None, None):
        ctx.shutdown(wait=True)


def shutdown_cancel_on_error(ctx: Executor, exc_info: ExcInfo):
    """
    Like `.shutdown_wait_on_error`, but invokes ``ctx.shutdown(wait=True,
    cancel_futures=True)``, and is instead designed for use in
    `~.SimpleEntryPoint.wrap_for_cancel_on_error` overrides.

    Only use this callback if your executor class accepts a ``cancel_futures``
    argument to its ``shutdown`` method.
    """
    if exc_info != (None, None, None):
        ctx.shutdown(wait=True, cancel_futures=True)


def shutdown_kill_on_error(ctx: Executor, exc_info: ExcInfo):
    """
    Like `.shutdown_wait_on_error`, but invokes ``ctx.shutdown(wait=True,
    kill_workers=True)``, and is instead designed for use in
    `~.SimpleEntryPoint.wrap_for_kill_on_error` overrides.

    Only use this callback if your executor class accepts a ``kill_workers``
    argument to its ``shutdown`` method.
    """
    if exc_info != (None, None, None):
        ctx.shutdown(wait=True, kill_workers=True)


class SimpleEntryPoint(ExecutorPluginEntryPoint):
    """
    Abstract subclass of `.ExecutorPluginEntryPoint` that works when the
    executor class' constructor is already friendly the default
    ``allowed_keywords`` property's implementation. Plugin authors simply need
    to override the following properties (can be done with simple attributes):

    - `.supports_serial`
    - `.supports_concurrent`
    - `.executor_module_name`
    - `.executor_class_name`
    - `.wrap_for_wait_on_error` (unless `.create` is overridden)
    - `.wrap_for_cancel_on_error` (unless `.create` is overridden)
    - `.wrap_for_kill_on_error` (unless `.create` is overridden)
    """

    @property
    def executor_module_name(self) -> str:
        """
        Fully qualified name of the module that hosts the executor class
        provided by this entry point. This class will lazily import the module
        so that `.is_available` checks can be made safely.

        This must be overridden for all subclasses.
        """
        raise NotImplementedError

    @property
    def executor_class_name(self) -> str:
        """
        Name of the executor class that is provided by this entry point. This
        should be just the class name itself, not including its module.

        This must be overridden for all subclasses.
        """
        raise NotImplementedError

    @property
    def extra_modules_required(self) -> Optional[Tuple[str]]:
        """
        If supplied, the default `.is_available` property will also check that
        these modules can be found by the importer.

        This must be overridden for all subclasses.
        """
        return ()

    @cached_property
    def is_available(self) -> bool:
        if extras := self.extra_modules_required:
            if isinstance(extras, str):
                # We'll let it slide if it looks like it's just a single module
                # name, but let's not start splitting strings here.
                if ("," in extras) or (":" in extras):
                    raise TypeError(
                        f"{self.extra_modules_required=!r} should "
                        f"be a tuple of strings."
                    )
                extras = (extras,)
            if not all(module_exists(name) for name in extras):
                return False
        return module_exists(self.executor_module_name)

    @cached_property
    def ExecutorCls(self) -> type:
        """
        Returns the lazily-imported executor class to use for this plugin. This
        uses the `.executor_module_name` and `.executor_class_name` attributes.

        If your plugin needs to use multiple classes or wrappers, consider
        directly subclassing `.ExecutorPluginEntryPoint` instead of using
        `.SimpleEntryPoint`.
        """
        if not self.is_available:
            # Safety check, in case the concrete subclass has a custom
            # is_available property.
            raise RuntimeError(
                "This entry point's module is unavailable and/or it has "
                "otherwise been disabled."
            )
        mod = importlib.import_module(self.executor_module_name)
        return getattr(mod, self.executor_class_name)

    @cached_property
    def allowed_keywords(self) -> Tuple[str]:
        return extract_keywords(self.ExecutorCls) | extract_keywords(type(self).create)

    @cached_property
    def default_max_workers(self):
        """
        When given ``max_workers=None``, `.create` will use this value as the
        maximum number of workers. Only applicable if `.supports_concurrent` is
        true. This base implementation uses the number of visible physical cores
        that are not blocked by the current CPU affinity mask. See
        `~pyrseus.core.sys.get_num_available_cores` for details.
        """
        return max(1, get_num_available_cores())

    def wrap_for_wait_on_error(self, ctx) -> Tuple[Executor, Optional[PreExitFunc]]:
        """
        Called by `.create` when its ``on_error`` argument is `.OnError.WAIT`,
        to ensure the returned context manager provides those semantics. Returns
        the same tuple type as `.ExecutorPluginEntryPoint.create`.

        This must be overridden for all subclasses, unless `.create` is
        overridden.
        """
        raise NotImplementedError

    def wrap_for_cancel_on_error(self, ctx) -> Tuple[Executor, Optional[PreExitFunc]]:
        """
        Called by `.create` when its ``on_error`` argument is
        `.OnError.CANCEL_FUTURES`, to ensure the returned context manager
        provides those semantics. Returns the same tuple type as
        `.ExecutorPluginEntryPoint.create`.

        This must be overridden for all subclasses, unless `.create` is
        overridden.
        """
        raise NotImplementedError

    def wrap_for_kill_on_error(self, ctx) -> Tuple[Executor, Optional[PreExitFunc]]:
        """
        Called by `.create` when its ``on_error`` argument is
        `.OnError.KILL_WORKERS`, to ensure the returned context manager provides
        those semantics. Returns the same tuple type as
        `.ExecutorPluginEntryPoint.create`.

        This must be overridden for all subclasses, unless `.create` is
        overridden.
        """
        raise NotImplementedError

    def create(
        self,
        max_workers: Optional[int] = None,
        *,
        on_error: OnErrorLike = OnError.KILL,
        **kwargs,
    ):
        """
        Creates an executor instance using this plugin.

        :param max_workers: the maximum number of workers to use. If `None`,
            this plugin will use `.default_max_workers` if it's a concurrent
            plugin.

        :param on_error: controls what `pyrseus.ctx.mgr.ExecutorCtx.__exit__`
            should do if an exception is passing through it. A pre-exit callback
            will be created if needed to provide the requested behavior. See
            `.get_standard_pre_exit_handler` for details.
        """
        if self.supports_concurrent and (max_workers is None):
            max_workers = self.default_max_workers
        on_error = OnError.get(on_error)

        ctx = self.ExecutorCls(max_workers, **kwargs)

        if on_error is OnError.WAIT:
            return self.wrap_for_wait_on_error(ctx)
        elif on_error is OnError.CANCEL_FUTURES:
            return self.wrap_for_cancel_on_error(ctx)
        elif on_error is OnError.KILL_WORKERS:
            return self.wrap_for_kill_on_error(ctx)
        else:
            raise ValueError(on_error)



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a simple factory for the 3rd
party `~mpi4py.futures.MPIPoolExecutor`, without overriding any of its defaults.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "mpi4py" works with normally-picklable tasks and return values
>>> skip_if_unavailable("mpi4py")
>>> with ExecutorCtx("mpi4py", 1) as exe:
...     assert exe.submit(os.getpid).result() != os.getpid()

>>> # But "mpi4py" doesn't use cloudpickle like "cpmpi4py" does,
>>> # so it can't handle fancy things like lambdas.
>>> skip_if_unavailable("mpi4py")
>>> with ExecutorCtx("mpi4py") as exe:
...     exe.submit(lambda: os.getpid()).result()
Traceback (most recent call last):
...
_pickle.PicklingError: Can't pickle ...

Note that unlike `multiprocessing`-based executors, the |mpi4py|_-based ones
only replicate the ``PYTHONPATH`` environment variable by default in workers,
not `sys.path`. If you have made local modifications to your `sys.path` list, be
sure to replicate them via the ``path`` argument that's passed to
`~.mpi4py.futures.MPIPoolExecutor`.

Plugin-specific Notes
---------------------

- *Common Use Cases:* Where one would otherwise use
  `~pyrseus.ctx.plugins.cpmpi4py` but where one wants to avoid depending on
  |cloudpickle|_.

- *Concurrency:* determined by the user's MPI configuration.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* |mpi4py|_

- *Underlying Executor:* `.mpi4py.futures.MPIPoolExecutor`

- *Default max_workers:* determined by the user's MPI configuration. Often this
  will require giving an explicit limit instead of relying on a pre-configured
  default.

- *Pickling:* `pickle`

- *OnError handling:* Fully supports `~pyrseus.ctx.api.OnError.WAIT` and
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` modes.
  `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode is automatically downgraded to
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES`.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

from functools import cached_property
from typing import Optional

from pyrseus.core.sys import module_exists
from pyrseus.ctx.api import ExecutorPluginEntryPoint, OnError, OnErrorLike
from pyrseus.ctx.simple import OnExitProxy


class EntryPoint(ExecutorPluginEntryPoint):
    supports_serial = False
    supports_concurrent = True

    @cached_property
    def is_available(self) -> bool:
        return module_exists("mpi4py.futures")

    allowed_keywords = {
        # Those supported by mpi4py. Unfortunately, their executor constructor
        # takes **kwargs, so we can't use metaprogramming like we do for other
        # plugins like process and loky.
        "python_exe",
        "python_args",
        "mpi_info",
        "globals",
        "main",
        "path",
        "wdir",
        "env",
    }

    def create(
        self,
        max_workers: Optional[int] = None,
        on_error: OnErrorLike = OnError.CANCEL_FUTURES,
        **kwargs,
    ):
        from mpi4py.futures import MPIPoolExecutor

        on_error = OnError.get(on_error)
        if on_error is OnError.KILL_WORKERS:
            # For now, demote KILL_WORKERS to CANCEL_FUTURES. MPIPoolExecutor
            # doesn't directly support a kill_workers mode. With the help of an
            # MPI expert, we'll be happy to consider adding a kill feature once
            # we know the right way to do it with MPI.
            on_error = OnError.CANCEL_FUTURES

        # Future work: if there's demand, add support for MPICommExecutor,
        # mpi4py's ThreadPoolExecutor, and/or mpi4py's ProcessPoolExecutor.
        ctx = MPIPoolExecutor(max_workers, **kwargs)

        if on_error is OnError.WAIT:
            return ctx, None
        elif on_error is OnError.CANCEL_FUTURES:
            exe = ctx.__enter__()  # just to be pedantic
            return OnExitProxy(exe, on_error=on_error, shutdown=True), None
        else:
            raise NotImplementedError(on_error)


ENTRY_POINT = EntryPoint()



"""
Tests some basic features of `pyrseus.interactive.get_executor`.
"""

import os
import sys
from threading import get_ident

from pyrseus.ctx.registry import skip_if_unavailable
from pyrseus.interactive import get_executor


def test_can_create_default_serial_exe():
    exe = get_executor(0)
    assert exe.submit(os.getpid).result() == os.getpid()
    assert exe.submit(get_ident).result() == get_ident()
    # Make sure there are no leaked references. 1 count is from our local
    # variable. The other count is from the argument to sys.getrefcount.
    assert sys.getrefcount(exe) == 2


def test_can_create_default_concurrent_exe():
    exe = get_executor(1)
    assert exe.submit(os.getpid).result() != os.getpid()
    # Make sure there are no leaked references. 1 count is from our local
    # variable. The other count is from the argument to sys.getrefcount.
    assert sys.getrefcount(exe) == 2


def test_loky_reuse():
    """
    Verifies that loky's reusable feature works properly, even with all of the
    wrapper layers we've added.
    """
    skip_if_unavailable("loky")
    exe0 = get_executor("loky", 1, reusable=True)
    exe1 = get_executor("loky", 1, reusable=True)
    assert exe0._obj._exe is exe1._obj._exe



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a wrapper around the 3rd
party |ipyparallel|_ executors that (a) provides a more standard executor-style
API, and (b) makes that library safer to use in production library code.

>>> import logging, os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "ipyparallel" can pickle fancy things like lambdas with cloudpickle.
>>> skip_if_unavailable("ipyparallel")
>>> with ExecutorCtx("ipyparallel", 1, log_level=logging.FATAL) as exe:
...     assert exe.__class__.__module__.startswith('ipyparallel')
...     assert exe.submit(lambda: os.getpid()).result() != os.getpid()

For more usage examples, see the `test_ipyparallel_plugin.py
<https://github.com/dalleyg/pyrseus/blob/main/tests/test_ipyparallel_plugin.py>`_
unit tests.

Plugin-specific Notes
---------------------

- *Common Use Cases:* for multi-host workloads, especially for for organizations
  that already have |ipyparallel|_ setup and wish to benefit from
  `~pyrseus.ctx.mgr.ExecutorCtx`'s ability to easily switch to other executors,
  especially the serial ones for light workloads and troubleshooting.

  - If your organization has subclassed or wrapped `.ipyparallel.Cluster` to
    help configure it better, consider deriving your own plugin from this basic
    one. See `.EntryPoint.cluster_module_name` for details.

  - *Non-recommended Use:* for basic single-host workloads, consider using
    `~pyrseus.ctx.plugins.process`, `~pyrseus.ctx.plugins.cpprocess`, or
    `~pyrseus.ctx.plugins.loky` instead of `~pyrseus.ctx.plugins.ipyparallel`.
    They require no setup beyond installation, they have fewer 3rd party package
    dependencies, their startup latency is one to two orders of magnitude lower,
    their workers use the same import environment as the primary process, and
    they have simpler interfaces.

  - *Non-recommended Use:* if you use many of |ipyparallel|_'s advanced features
    (other than configuration overrides), consider continuing to use it directly
    instead of via Pyrseus.

- *Concurrency:* determined by the user's |ipyparallel|_ configuration.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* as of ``ipyparallel`` 7.1.0 on Python 3.10,
  |ipyparallel|_ has 79 total transitive dependencies.

- *Underlying Executors:* miscellaneous wrappers around executors provided by
  |ipyparallel|_.

- *Default max_workers:* determined by the user's |ipyparallel|_ configuration.

- *Pickling:* |ipyparallel|_ is inconsistent about when it enables
  |cloudpickle|_ support. Since it's |ipyparallel|_'s preferred approach and
  since Pyrseus depends on |cloudpickle|_ anyway, this plugin unconditionally
  uses |cloudpickle|_ for pickling.

- *OnError handling:* Only directly supports implicit
  `~pyrseus.ctx.api.OnError.WAIT` semantics.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

import importlib
from contextlib import ExitStack
from functools import cached_property
from typing import Optional

from pyrseus.core.sys import module_exists
from pyrseus.ctx.api import ExecutorPluginEntryPoint, extract_keywords


class _ClusterCtx:
    """
    Automates the creation of a client and its executor from a cluster.
    """

    def __init__(self, cluster, targets):
        # Although we only actually need to keep track of the cluster and the
        # exit stack, we also maintain references to various other objects for
        # troubleshooting.
        self._es = ExitStack()
        self._cluster = cluster
        self._targets = targets
        self._client = None
        self._exe = None

    def __enter__(self):
        self._client = self._es.enter_context(self._cluster)
        self._client[:].use_cloudpickle()
        self._exe = self._client.executor(targets=self._targets)
        entered_exe = self._es.enter_context(self._exe)
        return entered_exe

    def __exit__(self, *exc_info):
        return self._es.__exit__(*exc_info)


class _ClientCtx:
    """
    Automates the creation of a client's executor, given a client.
    """

    def __init__(self, client, targets):
        self._es = ExitStack()
        self._client = client
        self._targets = targets
        self._exe = None

    def __enter__(self):
        self._client[:].use_cloudpickle()
        self._exe = self._client.executor(targets=self._targets)
        entered_exe = self._es.enter_context(self._exe)
        return entered_exe

    def __exit__(self, *exc_info):
        return self._es.__exit__(*exc_info)


class EntryPoint(ExecutorPluginEntryPoint):
    """
    Defines this plugin's entry point so that `~pyrseus.ctx.mgr.ExecutorCtx`
    knows how to use the plugin. This particular plugin is designed with further
    subclassing in mind. See `.cluster_class` and `.allowed_keywords` for
    details.
    """

    supports_serial = False
    supports_concurrent = True

    # NOTE: this cluster_module_name + cluster_class_name design is similar in
    # spirit to the way SimpleEntryPoint works. But we can't use it directly
    # since ipyparallel's interface doesn't directly provide an executor class;
    # users need to use the Cluster and Client classes first, in particular
    # ways, to then obtain an executor.

    cluster_module_name = "ipyparallel"
    """
    Gives the name of the module that contains the `.ipyparallel.Cluster` class
    or its factory that this plugin should use.

    If your organization has its own subclass or factory for clusters, then you
    may wish to create your own plugin that subclasses this one.

    - See :ref:`writingplugins` for general guidance on writing plugins.

    - Override this attribute to refer to your subclass or factory's module
      name.

    - Override the `.cluster_class_name` attribute to refer to your subclass or
      factory's name.

    - If your class uses traits the same way that `.ipyparallel.Cluster` does,
      that's all you should need to do. But if it's a factory function or if's a
      subclass it consumes constructor parameters differently, override
      `.allowed_keywords` as well. See
      `pyrseus.ctx.api.ExecutorPluginEntryPoint.allowed_keywords` for details.
    """

    cluster_class_name = "Cluster"
    """
    Name of the class or factory to import for creating `ipyparallel.Cluster`
    instances. See `.cluster_module_name` for details.
    """

    @cached_property
    def is_available(self) -> bool:
        return module_exists(self.cluster_module_name)

    @cached_property
    def ClusterCls(self) -> type:
        mod = importlib.import_module(self.cluster_module_name)
        return getattr(mod, self.cluster_class_name)

    @cached_property
    def allowed_keywords(self):
        # Merge the extra keywords that our create method consumes with the ones
        # that Cluster accepts. We remove "n" because our framework calls the
        # same thing "max_workers".
        ours = extract_keywords(EntryPoint.create)
        theirs = set(self.ClusterCls._traits)
        renamed = {"n"}
        return (ours | theirs) - renamed

    def create(
        self,
        max_workers: Optional[int] = None,
        *,
        client=None,
        cluster=None,
        targets=None,
        **cluster_kwargs,
    ):
        """
        Creates an `~concurrent.futures.Executor` that uses an |ipyparallel|_
        cluster and client pair as its backend.

        :param max_workers: an optional override to the maximum number of
            workers (engines in |ipyparallel|_ parlance) to use. This argument
            is silently ignored if either ``client`` or ``cluster`` is supplied.

        :param client: an optional `.ipyparallel.Client` object. If supplied,
            this client's default executor will be used by
            `~pyrseus.ctx.mgr.ExecutorCtx`. All other arguments except
            ``targets`` will be silently ignored.

        :param cluster: an optional `.ipyparallel.Cluster` object. Ignored if
            ``client`` is provided. Otherwise, if supplied, a client will be
            created from this cluster, and its default executor will be used by
            `~pyrseus.ctx.mgr.ExecutorCtx`. All other arguments except
            ``targets`` will be silently ignored.

        :param targets: an optional override that selects which |ipyparallel|_
            engines can be used by the executor. See
            `.ipyparallel.Client.executor` for details. If supplied, this
            argument is always honored.

        :param cluster_kwargs: additional keyword arguments passed to
            `.ipyparallel.Cluster`'s constructor if neither ``client`` nor
            ``cluster`` are provided.
        """
        if client is not None:
            return _ClientCtx(client, targets), None
        elif cluster is not None:
            return _ClusterCtx(cluster, targets), None
        else:
            cluster = self.ClusterCls(n=max_workers, **cluster_kwargs)
            return _ClusterCtx(cluster, targets), None


ENTRY_POINT = EntryPoint()



"""
Tests that basic ipyparallel-specific features are supported, if that library is
installed.
"""

import logging
import os
import platform
import time
from functools import cache

import psutil
import pytest

from pyrseus.ctx.mgr import ExecutorCtx
from pyrseus.ctx.registry import is_plugin_available

# Magic global variable that skips the whole test module if the plugin isn't
# installed.
#
# https://docs.pytest.org/en/latest/example/markers.html#scoped-marking
pytestmark = pytest.mark.skipif(
    not is_plugin_available("ipyparallel"), reason="ipyparallel is not available"
)

THIS_PROCESS = psutil.Process()


@cache
def get_worker_id():
    proc = psutil.Process()
    return (proc.pid, proc.create_time())


def get_inner_func():
    def inner_func():
        return 123

    return inner_func


def get_children():
    ret = set()
    for child in THIS_PROCESS.children():
        if "cov-report" in " ".join(child.cmdline()):
            # If running coverage tests, ignore its background processes that
            # come and go.
            continue
        try:
            # The pid and creation time are for uniqueness. The cmdline is for
            # troubleshooting.
            ret.add((child.pid, child.create_time(), " ".join(child.cmdline())))
        except psutil.NoSuchProcess:
            pass
    return ret


def test_simple_args():
    # Future work: troubleshoot why initial_children has an extra
    # multiprocessing.spawn.spawn_main process that disappears by the time we
    # hit exe_children. This only happens on macOS. Guess: ipyparallel is
    # changing the worker start mode so multiprocess kills a helper process for
    # the previous mode.
    if platform.system() == "Darwin":
        pytest.xfail("This test has spurious failures on macOS")

    # It's fine for some child processes to exist, as long as nothing else
    # changes that set within this test.
    initial_children = get_children()
    # Use ipyparallel to do some work.
    with ExecutorCtx("ipyparallel", 1) as exe:
        exe_children = get_children()
        # If something else in this test process is concurrently spawning or
        # reaping processes, it'll confuse our test.
        assert exe_children > initial_children
        # It should have started up 1 controller (fixed) and 1 engine (per our
        # request). This may need adjusting if they ever redesign their system,
        # e.g. add some kind of nanny process process of this one.
        assert len(exe_children) == len(initial_children) + 2
        # Run something trivial.
        assert exe.submit(os.getpid).result() != os.getpid()
        # We should still have the same children.
        assert exe_children == get_children()
    # Ensure it cleaned up after itself.
    assert initial_children == get_children()


@pytest.mark.slow
def test_cluster_provided():
    import ipyparallel as ipp

    initial_children = get_children()
    cluster = ipp.Cluster(n=1, log_level=logging.FATAL)
    try:
        # It should not auto-start anything yet. For this part, we give it a
        # little time since it takes a while to create and connect to its
        # subprocesses.
        time.sleep(2)
        assert initial_children == get_children()
        # Now trigger things normally.
        with ExecutorCtx("ipyparallel", cluster=cluster) as exe:
            exe_children = get_children()
            assert exe_children > initial_children
            assert len(exe_children) == len(initial_children) + 2
            assert exe.submit(os.getpid).result() != os.getpid()
            assert exe_children == get_children()
        assert initial_children == get_children()
    finally:
        cluster.stop_engines_sync()
        cluster.stop_controller_sync()
    assert initial_children == get_children()


@pytest.mark.slow
def test_client_provided_outside_ctx():
    import ipyparallel as ipp

    initial_children = get_children()
    cluster = ipp.Cluster(n=1, log_level=logging.FATAL)
    try:
        # It should not auto-start anything yet. For this part, we give it a
        # little time since it takes a while to create and connect to its
        # subprocesses.
        time.sleep(2)
        assert initial_children == get_children()
        # Once we synchronously start everything, then everything should proceed
        # as usual.
        client = cluster.start_and_connect_sync()
        client_children = get_children()
        assert client_children > initial_children
        assert len(client_children) == len(initial_children) + 2
        with ExecutorCtx("ipyparallel", client=client) as exe:
            assert client_children == get_children()
            assert exe.submit(os.getpid).result() != os.getpid()
            assert client_children == get_children()
        assert client_children == get_children()
    finally:
        cluster.stop_engines_sync()
        cluster.stop_controller_sync()
    assert initial_children == get_children()


@pytest.mark.slow
def test_client_provided_within_ctx():
    import ipyparallel as ipp

    initial_children = get_children()
    with ipp.Cluster(n=1, log_level=logging.FATAL) as client:
        client_children = get_children()
        assert client_children > initial_children
        assert len(client_children) == len(initial_children) + 2
        with ExecutorCtx("ipyparallel", client=client) as exe:
            assert client_children == get_children()
            assert exe.submit(os.getpid).result() != os.getpid()
            assert client_children == get_children()
        assert client_children == get_children()
    assert initial_children == get_children()


@pytest.mark.slow
def test_always_picklable_func():
    # If this crashes, everything will.
    with ExecutorCtx("ipyparallel", 1) as exe:
        exe.submit(os.getpid).result()


@pytest.mark.slow
def test_import_env_transferred_without_cloudpickle():
    # This test module sits outside the normal import hierarchy, so this will
    # crash unless the remote workers have replicated the import environment of
    # the main process.
    with ExecutorCtx("ipyparallel", 1) as exe:
        exe.submit(get_worker_id).result()


@pytest.mark.slow
def test_inner_func_pickling():
    # If this fails, it means that the pickler didn't figure out that the
    # function isn't at the global scope of its module.
    func = get_inner_func()

    with ExecutorCtx("ipyparallel", 1) as exe:
        assert exe.submit(func).result() == 123


@pytest.mark.slow
def test_lambda_pickling():
    # If this fails, it means the pickler can't handle lambdas properly.
    with ExecutorCtx("ipyparallel", 1) as exe:
        assert exe.submit(lambda: 123).result() == 123
        func = get_inner_func()
        assert exe.submit(lambda: func()).result() == 123
        assert exe.submit(lambda func=func: func()).result() == 123


@pytest.mark.slow
def test_double_picklability():
    # Here we verify that advanced pickling is available in both directions. We
    # send a lambda and get back a lambda.
    with ExecutorCtx("ipyparallel", 1) as exe:
        inner = lambda: 123  # NOQA
        outer = lambda: inner  # NOQA
        fut = exe.submit(outer)
        inner_ret = fut.result()
        ret = inner_ret()
        assert ret == 123



"""
Tests that `call_with_round_trip_pickling` and `try_pickle_round_trip` are able
to detect common picklability problems with exotic function types when they're
used in multi-process settings.
"""

import os
import pickle
import sys
from contextlib import contextmanager
from textwrap import dedent
from unittest import TestCase

import pytest

from pyrseus.core.pickle import call_with_round_trip_pickling, try_pickle_round_trip


def get_mul_2():
    def mul_2(x):
        return 2 * x

    return mul_2


def get_lambda():
    return lambda: 42


@contextmanager
def MainFuncCtx():
    main_mod = sys.modules["__main__"]
    main_globals = main_mod.__dict__
    name = "pyrseus_test_pickle_main_func"
    assert name not in main_globals  # check for improper cleanup of other tests
    exec(f"def {name}(): return 42", main_globals)
    yield main_globals[name]
    del main_globals[name]  # clean up after ourselves


@contextmanager
def MainClsCtx():
    main_mod = sys.modules["__main__"]
    main_globals = main_mod.__dict__
    name = "PyrseusTestPickleMainCls"
    assert name not in main_globals  # check for improper cleanup of other tests
    src = dedent(
        f"""
        class {name}:
            def mthd(self):
                return 42
        """
    ).strip()
    exec(src, main_globals)
    yield main_globals[name]
    del main_globals[name]  # clean up after ourselves


class TestTryPickleRoundTrip(TestCase):
    """
    Tests the basic functionality of `try_pickle_round_trip` with various tricky
    objects, including its ``hide_main`` behavior.
    """

    def test_normal_func(self):
        # Normal, importable functions work fine.
        assert try_pickle_round_trip(os.getpid) is os.getpid

    def test_main_func(self):
        with MainFuncCtx() as main_func:
            # We catch this case of a __main__ function being typically not
            # transferrable across processes.
            with pytest.raises(RuntimeError, match="Blocked attempt"):
                try_pickle_round_trip(main_func)

    def test_main_func_no_hide(self):
        with MainFuncCtx() as main_func:
            # The default picklers will happily pickle up a function defined in
            # __main__, since it doesn't know that the use may want to unpickle
            # it in a new process.
            assert try_pickle_round_trip(main_func, hide_main=False)() == 42

    def test_inner_func(self):
        # The built-in pickler already does a good job detecting functions that
        # are not in their owning module's globals dict. We report the native
        # exception.
        with pytest.raises(pickle.PicklingError):
            try_pickle_round_trip(get_mul_2())

    def test_lambda(self):
        # Similarly, the existing handling of lambdas is fine (given that pickle
        # can't handle them).
        with pytest.raises(pickle.PicklingError):
            try_pickle_round_trip(lambda: 42)

    def test_get_lambda(self):
        # But try_pickle_round_trip doesn't catch problems with unpicklable
        # return values. See TestCallWithRoundTripPickling for how
        # call_with_round_trip_pickling helps with that use case.
        assert try_pickle_round_trip(get_lambda)()() == 42

    def test_inst_method_of_normal_class(self):
        # Normal bound methods of importable classes work fine.
        data = {"a": 42}
        assert try_pickle_round_trip(data.get)("a") == 42

    def test_inst_method_of_main_class(self):
        with MainClsCtx() as MainCls:
            # We catch this case of a bound instance method for a class that was
            # defined in __main__.
            instance = MainCls()
            with pytest.raises(RuntimeError, match="Blocked attempt"):
                try_pickle_round_trip(instance.mthd)

    def test_inst_method_of_main_class_no_hide(self):
        with MainClsCtx() as MainCls:
            # The default picklers will happily pickle up a bound instance
            # method for a class that was defined in __main__, since it doesn't
            # know that the use may want to unpickle it in a new process.
            instance = MainCls()
            assert try_pickle_round_trip(instance.mthd, hide_main=False)() == 42


class TestCallWithRoundTripPickling(TestCase):
    """
    Tests the basic functionality of `call_with_round_trip_pickling`.
    This is very similar to `TestTryPickleRoundTrip`.
    """

    def test_normal_func(self):
        closure_args = (os.getpid, (), {})
        ret = call_with_round_trip_pickling(*closure_args)
        assert ret == os.getpid()

    def test_main_func(self):
        with MainFuncCtx() as main_func:
            closure_args = (main_func, (), {})
            with pytest.raises(RuntimeError, match="Blocked attempt"):
                call_with_round_trip_pickling(*closure_args)

    def test_main_func_no_hide(self):
        with MainFuncCtx() as main_func:
            closure_args = (main_func, (), {})
            ret = call_with_round_trip_pickling(*closure_args, hide_main=False)
            assert ret == 42

    def test_inner_func(self):
        closure_args = (get_mul_2, (), {})
        with pytest.raises(pickle.PicklingError):
            call_with_round_trip_pickling(*closure_args)

    def test_lambda(self):
        closure_args = (lambda: 42, (), {})
        with pytest.raises(pickle.PicklingError):
            call_with_round_trip_pickling(*closure_args)

    def test_get_lambda(self):
        # Here we can detect the picklability problem of get_lambda's return
        # value since we actually call the function and check the picklability
        # of the result.
        closure_args = (get_lambda, (), {})
        with pytest.raises(pickle.PicklingError):
            call_with_round_trip_pickling(*closure_args)

    def test_inst_method_of_normal_class(self):
        data = {"a": 42}
        closure_args = (data.get, ("a",), {})
        assert call_with_round_trip_pickling(*closure_args) == 42

    def test_inst_method_of_main_class(self):
        with MainClsCtx() as MainCls:
            instance = MainCls()
            closure_args = (instance.mthd, (), {})
            with pytest.raises(RuntimeError, match="Blocked attempt"):
                call_with_round_trip_pickling(*closure_args)

    def test_inst_method_of_main_class_no_hide(self):
        with MainClsCtx() as MainCls:
            instance = MainCls()
            closure_args = (instance.mthd, (), {})
            ret = call_with_round_trip_pickling(*closure_args, hide_main=False)
            assert ret == 42



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a simple serial executor,
`~pyrseus.executors.inline.InlineExecutor`, that captures exceptions in the
standard way.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "inline" doesn't pickle, so lambdas are fine.
>>> skip_if_unavailable("inline")
>>> with ExecutorCtx("inline") as exe:
...     assert exe.__class__.__name__ == 'InlineExecutor'
...     assert exe.submit(lambda: os.getpid()).result() == os.getpid()

Plugin-specific Notes
---------------------

- *Common Use Cases:*

  - Light workloads: this plugin is useful for avoiding concurrency overhead
    when running small batches of tasks. This lets developers avoid the
    alternative of rewriting all of their control flow to not use executors at
    all, just to get serial execution.

  - Troubleshooting: since tasks are executed immediately and within the same
    thread, tracing through the task code in a debugger is trivially easy.

- *Concurrency:* This is a non-concurrent, serial-only plugin. All tasks are
  immediately run in the same process and thread they were submitted in.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* This plugin has no 3rd party dependencies.
  Futhermore, it has minimal dependencies to other Pyrseus subpackages.

- *Underlying Executor:* This plugin uses
  `~pyrseus.executors.inline.InlineExecutor`. That executor has no external
  dependencies whatsoever.

- *Default max_workers:* Not applicable.

- *Pickling:* This plugin does not perform any pickling.

- *OnError handling:* Not applicable.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

from typing import Optional

from pyrseus.ctx.simple import SimpleEntryPoint


class EntryPoint(SimpleEntryPoint):
    supports_serial = True
    supports_concurrent = False
    executor_module_name = "pyrseus.executors.inline"
    executor_class_name = "InlineExecutor"

    def create(self, max_workers: Optional[int] = None, **kwargs):
        return self.ExecutorCls(**kwargs), None


ENTRY_POINT = EntryPoint()



from typing import Optional, Union

from .registry import filter_kwargs, get_entry_point

__all__ = ["ExecutorCtx"]


class ExecutorCtx:
    """
    Context manager that acts as a factory for many types of serial and
    concurrent `~.concurrent.futures.Executor` classes, where switching between
    executor plugins requires minimal user effort.

    Choosing the Plugin and Max Workers
    ===================================

    `.ExecutorCtx` accepts up to two positional arguments. They are used to
    choose which :ref:`plugin <plugins>` (and thus underlying executor) will be
    used and what its ``max_workers`` argument will be. All other constructor
    arguments will be passed to the plugin.

    No Positional Arguments
    -----------------------

    With no positional arguments, `.ExecutorCtx` will default to creating a
    multiprocess executor, with one worker per available CPU core, respecting
    the current CPU affinity mask (see
    `~pyrseus.core.sys.get_num_available_cores`).

        >>> with ExecutorCtx() as exe:
        ...    print(exe.submit(sum, [1, 2, 3, 4]).result())
        10

    One Positional Integer Argument
    -------------------------------

    If one positional integer argument is supplied, `.ExecutorCtx` will either
    use a special serial executor (if 0):

        >>> import os
        >>> with ExecutorCtx(0) as exe:
        ...     assert exe.submit(os.getpid).result() == os.getpid()  # serial

    or its default concurrent executor (if positive):

        >>> with ExecutorCtx(1) as exe:
        ...     assert exe.submit(os.getpid).result() != os.getpid()  # concurrent

    One Positional String Argument
    ------------------------------

    If one positional string argument is supplied, it names a specific
    `.ExecutorCtx` plugin that should be used. That plugin's default
    ``max_workers`` will be used.

        >>> with ExecutorCtx("inline") as exe:
        ...     # "inline" is a serial plugin that immediately runs tasks
        ...     # in the same process and thread.
        ...     assert exe.submit(os.getpid).result() == os.getpid()

    Two Positional Arguments
    ------------------------

    If two positional arguments are supplied, the first must be the plugin name,
    and the second the ``max_workers`` to use with it:

        >>> with ExecutorCtx("process", 1) as exe:
        ...     futs = [exe.submit(os.getpid) for _ in range(10)]
        ...     some_worker_pids = {fut.result() for fut in futs}
        ...     assert len(some_worker_pids) == 1
        ...     assert all(
        ...         worker_pid != os.getpid()
        ...         for worker_pid in some_worker_pids
        ...     )

    Extra Arguments
    ===============

    After the plugin and/or ``max_workers`` arguments, all others are
    keyword-only. See `~pyrseus.ctx.registry.get_all_allowed_keywords`,
    `~pyrseus.ctx.registry.get_keywords_by_plugin`, and
    `~pyrseus.ctx.registry.get_keywords_for_plugin` for the full list of
    supported keywords based on which plugins are installed on your system.

    Irrelevant Keywords are Auto-ignored
    ------------------------------------

    Keyword arguments that are not supported by the selected plugin are silently
    ignored. This makes it easier for users to switch between executor types by
    just changing the positional arguments and not needing to update all of the
    keyword arguments.

    For example, suppose you're troubleshooting a pickling problem with the
    `~pyrseus.ctx.plugins.mpi4py` plugin, where you've also added several
    plugin-specific keyword arguments to the `.ExecutorCtx` constructor call:

        >>> # Uh oh, we have a pickling problem!
        >>> from pyrseus.ctx.registry import skip_if_unavailable
        >>> skip_if_unavailable("mpi4py")
        >>> with ExecutorCtx(
        ...     "mpi4py",
        ...     main=True, path=['some/lib', 'another/lib'] # (cp)mpi4py-specific kwargs
        ... ) as exe:
        ...     print(exe.submit(lambda: 42).result())
        Traceback (most recent call last):
        ...
        _pickle.PicklingError: Can't pickle ...

    A good way to troubleshoot these types of issues is with the
    `~pyrseus.ctx.plugins.pinline` and/or `~pyrseus.ctx.plugins.cpinline`
    plugins. We can switch to either of those by just changing the first
    positional argument and leaving the other |mpi4py|_-specific arguments
    there. In this case, we see that the problem is that we need a
    |cloudpickle|_-enabled plugin like `~pyrseus.ctx.plugins.cpmpi4py` instead
    of the base `~pyrseus.ctx.plugins.mpi4py`.

        >>> # We can reproduce it with the pinline plugin. If we have a debugger
        >>> # like pdb or ipdb, we could trace into the code since this plugin
        >>> # runs its tasks immediately and in the same process and thread
        >>> # (debugging not shown here).
        >>> #
        >>> # NOTE: the pinline plugin doesn't take `main` and `path` arguments,
        >>> # but ExecutorCtx knows this so it suppresses them.
        >>> skip_if_unavailable("pinline")
        >>> with ExecutorCtx(
        ...     "pinline",
        ...     main=True, path=['some/lib', 'another/lib'] # auto-ignored
        ... ) as exe:
        ...     print(exe.submit(lambda: 42).result())
        Traceback (most recent call last):
        ...
        _pickle.PicklingError: Can't pickle ...

        >>> # Suppose we realize that we should be using a cloudpickle-enabled
        >>> # plugin. We try cpinline and see that it works.
        >>> skip_if_unavailable("cpinline")
        >>> with ExecutorCtx(
        ...     "cpinline",
        ...     main=True, path=['some/lib', 'another/lib'] # auto-ignored
        ... ) as exe:
        ...     print(exe.submit(lambda: 42).result())
        42

        >>> # Then we switch to cpmpi4py and see that it works too.
        >>> skip_if_unavailable("cpmpi4py")
        >>> with ExecutorCtx(
        ...     "cpmpi4py",
        ...     main=True, path=['some/lib', 'another/lib'] # (cp)mpi4py-specific kwargs
        ... ) as exe:
        ...     print(exe.submit(lambda: 42).result())
        42

    Invalid Keywords Give an Exception
    ----------------------------------

    Although `.ExecutorCtx` ignores keywords that are irrelevant for the
    currently selected plugin, it does check for keywords that aren't relevant
    for any plugins. This helps catch user mistakes.

        >>> with ExecutorCtx(this_is_not_a_valid_keyword="for_any_plugin"):
        ...     pass
        Traceback (most recent call last):
        ...
        TypeError: Keyword 'this_is_not_a_valid_keyword' is not accepted by any of the registered plugins:
            plugins: [...'inline'...]
            all allowed keywords: [...]

    Changing the Default Plugins
    ============================

    Pyrseus' default plugins are `~pyrseus.ctx.plugins.inline` when
    ``max_workers=0`` and `~pyrseus.ctx.plugins.process` otherwise. To change
    either of these defaults, use
    `~pyrseus.ctx.registry.SetDefaultExecutorPluginCtx` or
    `~pyrseus.ctx.registry.set_default_executor_plugin`.

    Creating New Plugins
    ====================

    See :ref:`writingplugins`.
    """

    def __init__(
        self,
        plugin_or_max_workers: Union[str, int, None] = None,
        /,
        max_workers: Optional[int] = None,
        **kwargs,
    ):
        plugin_name, max_workers, entry_point = get_entry_point(
            plugin_or_max_workers, max_workers
        )
        self._relevant_kwargs = filter_kwargs(entry_point, kwargs)

        # Record additional metadata for troubleshooting.
        self._plugin_name = plugin_name
        self._max_workers = max_workers
        self._entry_point = entry_point
        self._full_kwargs = kwargs

        # Invoke the factory and save the results.
        self._exe, self._pre_exit = entry_point.create(
            self._max_workers, **self._relevant_kwargs
        )

    def __enter__(self):
        """
        Enters the context of the underlying `~concurrent.futures.Executor`.
        """
        return self._exe.__enter__()

    def __exit__(self, *exc_info):
        if self._pre_exit is not None:
            self._pre_exit(self._exe, exc_info)
        return self._exe.__exit__(*exc_info)



import inspect
import os
import sys
import threading
from concurrent.futures import ProcessPoolExecutor
from unittest import TestCase
from unittest.mock import Mock

from pyrseus.ctx.mgr import ExecutorCtx

PLUGIN_NAME = "thread"


def get_worker_id():
    return threading.get_ident()


def set_tls_worker_id(tls):
    tls.worker_id = get_worker_id()


def get_tls_worker_id(tls):
    return tls.worker_id


class TestThreadPoolKwargs(TestCase):
    POOL_SIG = inspect.signature(ProcessPoolExecutor)

    def test_ignored(self):
        """
        Ensure that various the parameters that `.ExecutorCtx` is supposed to
        ignore for this executor type are in fact ignored.
        """
        # For all parameters that are supposed to be ignored, pass in a poisoned
        # argument: Mock() is sufficiently incompatible for us.
        mock = Mock()
        kw = {"mp_context": mock}
        if sys.version_info >= (3, 11):
            kw["max_tasks_per_child"] = mock
        with ExecutorCtx(PLUGIN_NAME, 1, **kw) as exe:
            assert exe.submit(os.getpid).result() == os.getpid()
            assert exe.submit(get_worker_id).result() != get_worker_id()
        assert not mock.called
        assert len(mock.method_calls) == 0
        assert len(mock.mock_calls) == 0

    def test_bad_kwarg_not_ignored(self):
        with self.assertRaises(TypeError):
            with ExecutorCtx(PLUGIN_NAME, this_is_not_a_valid_kwarg_name=123):
                pass

    def test_initializer(self):
        tls = threading.local()
        # First verify that tls.worker_id isn't set in the main thread.
        with self.assertRaises(AttributeError):
            tls.worker_id
        # If we don't run the initializer, then the code fails in threads.
        with ExecutorCtx(PLUGIN_NAME, 1) as exe:
            fut = exe.submit(get_tls_worker_id, tls)
            with self.assertRaises(AttributeError):
                fut.result()
        # But we can initialize thread local storage, and then it works fine.
        with ExecutorCtx(
            PLUGIN_NAME, 1, initializer=set_tls_worker_id, initargs=(tls,)
        ) as exe:
            assert exe.submit(get_tls_worker_id, tls).result() != get_worker_id()

    def test_thread_name_prefix(self):
        PREFIX = "TestThread-PYRSEUS-CTX-TTPK-TTNP"

        def count_threads_with_prefix():
            count = 0
            for thread in threading.enumerate():
                if thread.name.startswith(PREFIX):
                    count += 1
            return count

        assert count_threads_with_prefix() == 0
        with ExecutorCtx(PLUGIN_NAME, 1, thread_name_prefix=PREFIX) as exe:
            fut = exe.submit(get_worker_id)
            assert count_threads_with_prefix() == 1
            assert fut.result() != get_worker_id()
            assert count_threads_with_prefix() == 1
        assert count_threads_with_prefix() == 0



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that is a variant of the
`~pyrseus.ctx.plugins.inline` one, but this one tests the picklability of all
submitted tasks and their return values, using |cloudpickle|_. This plugin uses
`~pyrseus.executors.cpinline.CpInlineExecutor`.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "cpinline" tests picklability with cloudpickle, e.g. for lambdas
>>> skip_if_unavailable("cpinline")
>>> with ExecutorCtx("cpinline") as exe:
...     assert exe.__class__.__name__ == 'CpInlineExecutor'
...     assert exe.submit(lambda: os.getpid()).result() == os.getpid()

Plugin-specific Notes
---------------------

- *Common Use Cases:* serially troubleshooting pickling problems for
  multi-process plugins that use |cloudpickle|_ for serialization.

- *Concurrency:* This is a non-concurrent, serial-only plugin. All tasks are
  immediately run in the same process and thread they were submitted in.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* |cloudpickle|_

- *Underlying Executor:* `~pyrseus.executors.cpinline.CpInlineExecutor`

- *Default max_workers:* Not applicable.

- *Pickling:* This plugin takes extra time to pickle and unpickle all tasks and
  their results. If you aren't troubleshooting such issues and prefer lower
  overhead, consider using the `~pyrseus.ctx.plugins.inline` plugin instead.

- *OnError handling:* Not applicable.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

from typing import Optional

from pyrseus.ctx.simple import SimpleEntryPoint


class EntryPoint(SimpleEntryPoint):
    supports_serial = True
    supports_concurrent = False
    executor_module_name = "pyrseus.executors.cpinline"
    executor_class_name = "CpInlineExecutor"
    extra_modules_required = ("cloudpickle",)

    def create(self, max_workers: Optional[int] = None, **kwargs):
        return self.ExecutorCls(**kwargs), None


ENTRY_POINT = EntryPoint()



from unittest import TestCase

from pyrseus.executors.pnocatch import PNoCatchExecutor

Cls = PNoCatchExecutor


def picklable():
    return 1


def make_lambda():
    return lambda: 2


def make_double_unpicklable():
    return lambda: lambda: 3


class TestPickling(TestCase):
    def test_picklable(self):
        with Cls() as exe:
            fut = exe.submit(picklable)
            ret = fut.result()
            assert ret == 1

    def test_lambda(self):
        with Cls() as exe:
            lambd = make_lambda()
            with self.assertRaisesRegex(Exception, "Can't pickle"):
                exe.submit(lambd)

    def test_return_lambda(self):
        with Cls() as exe:
            with self.assertRaisesRegex(Exception, "Can't pickle"):
                exe.submit(make_lambda)

    def test_double_unpicklable(self):
        with Cls() as exe:
            func = make_double_unpicklable()
            with self.assertRaisesRegex(Exception, "Can't pickle"):
                exe.submit(func)



"""
Provides a variant of `InlineExecutor` that tests the picklability of all
submitted tasks and their return values, using |cloudpickle|_.
"""

from functools import partial

import cloudpickle

from pyrseus.core.pickle import call_with_round_trip_pickling
from pyrseus.executors.inline import InlineExecutor


class CpInlineExecutor(InlineExecutor):
    """
    Pickle-testing variant of `.InlineExecutor`, using the 3rd party
    |cloudpickle|_ module.

    This variant pickles each submitted task and the task's results, using the
    |cloudpickle|_ module that understands things like lambdas. This is
    primarily useful for troubleshooting pickling problems occurring in
    multi-process executors, by performing the pickling and unpickling locally.

    Consider this lambda.

        >>> import pickle
        >>> needs_cloudpickle = lambda: "this works if using cloudpickle for pickling"

    It can't be pickled with `pickle`.

        >>> pickle.dumps(needs_cloudpickle, -1)
        Traceback (most recent call last):
        ...
        _pickle.PicklingError: Can't pickle ...

    But it can be pickled with |cloudpickle|_.

        >>> print(pickle.loads(cloudpickle.dumps(needs_cloudpickle, -1))())
        this works if using cloudpickle for pickling

    Since this class uses |cloudpickle|_ for pickling, it works fine.

        >>> with CpInlineExecutor() as exe:
        ...     fut = exe.submit(needs_cloudpickle)
        ...     print(fut.result())
        this works if using cloudpickle for pickling

    See :doc:`../plugins` for a list of related executors.
    """

    def submit(self, fcn, /, *args, **kwargs):
        testing_closure = partial(
            call_with_round_trip_pickling,
            fcn,
            args,
            kwargs,
            dumps=cloudpickle.dumps,
            loads=cloudpickle.loads,
            # __main__ dependency problems are very rare with cloudpickle, so
            # for now we disable the temporary sys.modules patching.
            hide_main=False,
        )
        return super().submit(testing_closure)



"""
System-level utilities for the `pyrseus` package.
"""

import importlib
import multiprocessing
import signal
from contextlib import contextmanager
from functools import cache
from typing import Optional

import psutil

LOGICAL_CORES = psutil.cpu_count(logical=True)
PHYSICAL_CORES = psutil.cpu_count(logical=False)  # None on some platforms


def get_num_available_cores(pid: Optional[int] = None, physical: bool = False):
    """
    Returns the number of cores that are available to the given process for
    scheduling threads, respecting the CPU affinity mask when possible. Think of
    this as a better version of `multiprocessing.cpu_count`.

    .. note:

       If you are calling this in code that already depends on the |loky|_
       package, consider using `.loky.backend.context.cpu_count` instead of this
       function. It is even more sophisticated.

    :param pid: process ID to test, on systems that have affinity masks.
        Defaults to the current process.

    :param physical: whether to try only counting the number of physical cores
        instead of logical cores. Silently ignored on platforms that don't
        support querying the physical core count.
    """
    if hasattr(psutil.Process, "cpu_affinity"):
        # Use psutil on any platform that supports it (Linux, Windows, FreeBSD).
        proc = psutil.Process(pid)
        logical_available = len(proc.cpu_affinity())
        if physical:
            if (PHYSICAL_CORES is not None) and (PHYSICAL_CORES >= 0):
                # On most platforms, we can do the right thing.
                return int((PHYSICAL_CORES * logical_available) // LOGICAL_CORES)
            else:
                # On OpenBSD and NetBSD, physical is always None, so give up and
                # just assume it's okay to return the number of available logical
                # cores.
                return logical_available
        else:
            return logical_available
    else:
        # Fallback to just counting the number of visible cores on platforms
        # that don't have Process.cpu_affinity, like macOS.
        return multiprocessing.cpu_count()


@cache
def is_mp_start_method_supported(start_method: str) -> bool:
    """
    Returns whether the given ``start_method`` is supported by the
    `multiprocessing` library, in the current process.
    """
    import multiprocessing as mp

    try:
        mp.get_context(start_method)
    except ValueError:
        return False
    else:
        return True


def module_exists(name):
    """
    Tells whether the given Python module or package exists and looks
    importable, without actually importing it.
    """
    try:
        return importlib.util.find_spec(name) is not None
    except ModuleNotFoundError:
        return False


@contextmanager
def SignalHandlerCtx(signum, handler):
    """
    Safely makes a temporary change to one signal handler.
    """
    unset = object()
    old_handler = unset
    try:
        old_handler = signal.signal(signum, handler)
        yield
    finally:
        if old_handler is not unset:
            signal.signal(signum, old_handler)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown] editable=true slideshow={"slide_type": ""} tags=["pyflyby-cell"]
# # Debugging Tasks with Pyrseus
#
# In this notebook, we show how Pyrseus can help troubleshoot problems with
# tasks, especially problems with pickling.

# %% [markdown]
# ## Setup
#
# Before running this notebook, make sure everything it depends on is installed:
#
#     # Modify the docs/requirements.txt path if you're running this
#     # command from anything except the repository's root directory.
#     python -m pip install -r docs/requirements.txt
#
# These are some imports that we'll use throughout this notebook:

# %%
import sys

sys.path.append("../../../src")  # assume we're running it from a Pyrseus source clone
import pickle
import random
from multiprocessing import get_context

import cloudpickle

from pyrseus.core.pickle import call_with_round_trip_pickling, try_pickle_round_trip
from pyrseus.ctx.mgr import ExecutorCtx

# %% [markdown]
# And here's a simple custom function we'll be experimenting with.
# It works like `sorted`, but it uses the (slow) Selection Sort algorithm.


# %%
def selection_sort(data):
    """
    Simple selection sort.

    Adapted from: https://en.wikipedia.org/wiki/Selection_sort
    """
    # Make a shallow copy of the data so that this can be a
    # non-mutating function.
    ret = list(data)

    # Now perform the selection sort.
    for i in range(len(data)):
        j_min = i
        for j in range(i + 1, len(data)):
            if ret[j] < ret[j_min]:
                j_min = j
        if j_min != i:
            ret[j_min], ret[i] = ret[i], ret[j_min]

    return ret


# %% [markdown]
# ## Non-executor Usage
#
# Let's first try out the function by calling it directly with a few hand-crafted test cases.

# %%
for data in (
    (),
    (1,),
    (1, 2),
    (1, 2, 3),
    (1, 2, 3, 4),
    (1, 2, 3, 4, 5),
    (1, 5, 2, 4, 3),
    (5, 4, 3, 2, 1),
    (4, 3, 2, 1),
    (3, 2, 1),
):
    expected = sorted(data)
    actual = selection_sort(data)
    assert actual == expected, (data, actual, expected)
    print(f"{str(data):<15s} -> {actual}")


# %% [markdown]
# Let's also create a randomized test helper function and run it
# on a few different inputs.


# %%
def sorting_test_with_big_random_list(seed, n=1000, min_int=0, max_int=500):
    # Make this test repeatable.
    random.seed(seed)
    # Generate some random data.
    data = [random.randint(min_int, max_int) for _ in range(n)]
    # Sort with our method.
    actual = selection_sort(data)
    # Sort with a known good implementation.
    expected = sorted(data)
    # Tell whether the two match.
    return actual == expected
    if actual != expected:
        raise ValueError(f"Results for: {seed=}, {n=}, {min_int=}, {max_int=}")


assert sorting_test_with_big_random_list(0)
assert sorting_test_with_big_random_list(42)

# %% [markdown]
# ## Failures with ProcessPoolExecutor
#
# Now suppose we want to run that test helper many times in parallel, using
# `ProcessPoolExecutor`. Unfortunately, we quickly run into trouble.
# Depending on your Python version, your platform, and exactly what was
# submitted, this will result in at least one of the following:
#  - dead workers (undesirable),
#  - workers printing messages to stderr (undesirable),
#  - `BrokenProcessPool` exceptions (a symptom),
#  - exceptions talking about `'__main__'` (a symptom),
#  - exceptions talking about pickling (the real problem), and/or
#  - exceptions talking about unpickling (a symptom).
#
# For the sake of this notebook, we force it to use the most widely supported
# pool type (`"spawn"`). At least on Python 3.10, this type also results in
# the most verbose and confusing output.

# %% editable=true slideshow={"slide_type": ""} tags=["raises-exception"]
# In this cell, we first encounter a problem running our function
# in some multiprocessing workers.
try:
    print("This test should print out many stderr lines from the workers, but succeed.")
    sys.stdout.flush()

    with ExecutorCtx("process", 4, mp_context=get_context("spawn")) as exe:
        futs = [
            exe.submit(sorting_test_with_big_random_list, seed) for seed in range(25)
        ]
        for seed, fut in enumerate(futs):
            if not fut.result():
                print(f"Seed {seed} failed.")

except Exception as ex:
    print("CAUGHT EXCEPTION (expected):", ex)
else:
    raise RuntimeError("An exception should have been thrown.")

# %% [markdown]
# ## Troubleshooting with Serial Executors
#
# In cases where the exception message doesn't make it clear to the user
# what to do, a common strategy is to run the code serially. Fortunately,
# `ExecutorCtx` makes this trivial to do: we just need to change the plugin
# name (or the `max_workers` argument if the plugin name is being omitted).

# %%
# In this cell, we've failed to replicate the problem.
#
# Notice that all we had to do was change "process"
# to "inline" to try this out. We didn't need to remove
# the 4 or mp_context arguments.
with ExecutorCtx("inline", 4, mp_context=get_context("spawn")) as exe:
    futs = [exe.submit(sorting_test_with_big_random_list, seed) for seed in range(25)]
    for seed, fut in enumerate(futs):
        if not fut.result():
            print(f"Seed {seed} failed.")

# %% [markdown]
# Unfortunately, the above snippet doesn't reproduce the problem. Let's
# assume this led to us doing some more experiments and/or web searches,
# making us think this could be related to pickling and/or unpickling.
#
# At this point, we may try using `PInlineExecutor`, since it advertises
# itself as a tool for troubleshooting pickling problems. And indeed we
# now have a reproducer.

# %%
# In this cell, we have successfully replicated the problem with a serial executor.
#
# Notice that all we had to do was change "inline"
# to "pinline" to try this out.
try:
    with ExecutorCtx("pinline", 4, mp_context=get_context("spawn")) as exe:
        futs = [
            exe.submit(sorting_test_with_big_random_list, seed) for seed in range(25)
        ]
        for seed, fut in enumerate(futs):
            if not fut.result():
                print(f"Seed {seed} failed.")

except Exception as ex:
    print("CAUGHT EXCEPTION (expected):", ex)

# %% [markdown]
# Additionally, this test was done with the pure Python pickler,
# so we could even trace into it with ``ipdb`` if we want.

# %%
# If you'd like to try debugging it yourself, then
#  (a) remove the try-except wrapper around the previous cell,
#  (b) uncomment the %debug line from this cell, and
#  (c) run both cells, one at a time.

# # %debug

# %% [markdown]
# ## Testing Cloudpickle Serially
#
# At this point, we have figured out it's a picklability problem. Hopefully we've also
# figured out that the problem is that we're using a function that's defined in
# ``__main__`` instead of in an imported module. Additionally, we've hopefully heard
# about the `cloudpickle` library being a solution to this kind of problem.
#
# We could test this last hypothesis in a few ways. First, let's verify whether
# `cloudpickle` works at all on our function. It does:

# %%
# cloudpickle says it can handle our function, so we have a chance.
pickled = cloudpickle.dumps(sorting_test_with_big_random_list, -1)
reconstructed = cloudpickle.loads(pickled)
assert reconstructed(0)

# %% [markdown]
# That said, we probably shouldn't trust `cloudpickle` that much since `pickle` thought
# it could handle our function too (and technically it can, but only when we don't
# send the pickled bytestring to another process for unpickling).
#
# Fortunately, Pyrseus ships with a simple test function that simulates this situation.
# First let's show that we can replicate the problem with it when using `pickle`.

# %%
# First, make sure that try_pickle_round_trip can replicate our problem when using pickle.
try:
    try_pickle_round_trip(
        sorting_test_with_big_random_list,
        dumps=pickle.dumps,
        loads=pickle.loads,
        hide_main=True,  # The default is true. We include it here for emphasis.
    )
except Exception:
    print("try_pickle_round_trip successfully replicated the problem with pickle.")
else:
    raise RuntimeError("try_pickle_round_trip failed to replicate the problem.")

# %% [markdown]
# Now let's try using that function to see it thinks `cloudpickle` will fix our problems.

# %%
# Indeed, try_pickle_round_trip tells us that if we use cloudpickle,
# then our pickling problems will likely go away.
reconstructed = try_pickle_round_trip(
    sorting_test_with_big_random_list,
    dumps=cloudpickle.dumps,
    loads=cloudpickle.loads,
)
assert reconstructed(0)

# %% [markdown]
# We might also try using an even more complete tester that internally:
# - runs `try_pickle_round_trip` on the function (similar to above),
# - calls the function (so we see if the call itself is a problem), and
# - runs `try_pickle_round_trip` on the function result (in case there's a picklability problem with it).

# %%
# call_with_round_trip_pickling also thinks that everything's good if we
# switch to using cloudpickle.
assert call_with_round_trip_pickling(
    sorting_test_with_big_random_list,
    args=(0,),
    kwargs={},
    dumps=cloudpickle.dumps,
    loads=cloudpickle.loads,
)

# %% [markdown]
# ## Trying a Cloudpickle-enabled Executor
#
# So now let's try some `cloudpickle`-enabled executors.
#
# First, we see that `CpProcessPoolExecutor` works fine. It's just
# a thin wrapper around `ProcessPoolExecutor` that uses `cloudpickle`
# for pickling tasks and their results.

# %%
# CpProcessPoolExecutor works! Also, all we had to do was change the
# plugin name to "cpprocess".
with ExecutorCtx("cpprocess", 4, mp_context=get_context("spawn")) as exe:
    futs = [exe.submit(sorting_test_with_big_random_list, seed) for seed in range(25)]
    for seed, fut in enumerate(futs):
        if not fut.result():
            print(f"Seed {seed} failed.")

# %% [markdown]
# We might also try a fancier one like `loky`'s. Their executor is
# a from-scratch rewrite of `ProcessPoolExecutor`, with built-in
# `cloudpickle` support, and various robustness improvements over the
# built-in one. It works fine too.

# %%
# loky's ProcessPoolExecutor also works. Similar to before, we only had to
# change the plugin name to "loky".
with ExecutorCtx("loky", 4, mp_context=get_context("spawn")) as exe:
    futs = [exe.submit(sorting_test_with_big_random_list, seed) for seed in range(25)]
    for seed, fut in enumerate(futs):
        if not fut.result():
            print(f"Seed {seed} failed.")

# %% [markdown]
# Now, we're done debugging. We know that we just need to make sure
# we use a `cloudpickle`-enabled plugin like `"cpprocess"` or `"loky"`.



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a simple serial executor,
`~pyrseus.executors.nocatch.NoCatchExecutor`, that does *not* catch exceptions
in submitted futures.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "nocatch" doesn't catch exceptions at submit time
>>> skip_if_unavailable("nocatch")
>>> def raises():
...     raise RuntimeError(
...         "I am an exception, passed out at submit time, "
...         "instead of held until fut.result() was called."
...     )
>>> with ExecutorCtx("nocatch") as exe:
...     assert exe.__class__.__name__ == 'NoCatchExecutor'
...     exe.submit(raises)                    # <-- NOTE: no fut.result() needed
Traceback (most recent call last):
...
RuntimeError: I am an exception, passed out at submit time,
instead of held until fut.result() was called.

>>> # "nocatch" doesn't pickle, so lambdas are fine.
>>> skip_if_unavailable("nocatch")
>>> with ExecutorCtx("nocatch") as exe:
...     assert exe.submit(lambda: os.getpid()).result() == os.getpid()

Plugin-specific Notes
---------------------

- *Common Use Cases:* for troubleshooting, as a fail-fast variant of the
  `~pyrseus.ctx.plugins.inline`.

- *Concurrency:* This is a non-concurrent, serial-only plugin. All tasks are
  immediately run in the same process and thread they were submitted in.

- *Exceptions:* This plugin has *non-standard* exception-handling semantics: no
  task exceptions are caught and captured in their futures. Exceptions are
  propagated out immediately.

- *3rd Party Dependencies:* This plugin has no 3rd party dependencies.
  Futhermore, it has minimal dependencies to other Pyrseus subpackages.

- *Underlying Executor:* This plugin uses
  `~pyrseus.executors.nocatch.NoCatchExecutor`. That executor has no external
  dependencies whatsoever.

- *Default max_workers:* Not applicable.

- *Pickling:* This plugin does not perform any pickling.

- *OnError handling:* Not applicable.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

from typing import Optional

from pyrseus.ctx.simple import SimpleEntryPoint


class EntryPoint(SimpleEntryPoint):
    supports_serial = True
    supports_concurrent = False
    executor_module_name = "pyrseus.executors.nocatch"
    executor_class_name = "NoCatchExecutor"

    def create(self, max_workers: Optional[int] = None, **kwargs):
        return self.ExecutorCls(**kwargs), None


ENTRY_POINT = EntryPoint()



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a wrapper around the
built-in `~concurrent.futures.ProcessPoolExecutor` that (a) uses |cloudpickle|_
instead of `pickle` for pickling tasks and their return values, and (b) respects
the CPU affinity mask when choosing the default thread count.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "cpprocess" can pickle fancy things like lambdas with cloudpickle
>>> skip_if_unavailable("cpprocess")
>>> with ExecutorCtx("cpprocess", 1) as exe:
...     assert exe.submit(lambda: os.getpid()).result() != os.getpid()

Plugin-specific Notes
---------------------

- *Common Use Cases:* For the same use cases as `~pyrseus.ctx.plugins.process`,
  but when users wish to use |cloudpickle|_ instead of `pickle` for serializing
  tasks and their results.

- *Concurrency:* Each worker runs in its own process.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* |cloudpickle|_

- *Underlying Executor:* `concurrent.futures.ProcessPoolExecutor`

- *Default max_workers*:* Uses `~pyrseus.core.sys.get_num_available_cores`
  instead of `multiprocessing.cpu_count`, respecting the CPU affinity mask when
  possible. Currently is unaware of cgroups constraints.

- *Pickling:* |cloudpickle|_

- *OnError handling:* Fully supports `~pyrseus.ctx.api.OnError.WAIT` and
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` modes.
  `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode is automatically downgraded to
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES`.

  - By default, `~concurrent.futures.ProcessPoolExecutor`, pre-queues one extra
    task, making it uncancellable; so in
    `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` mode, that pre-queued task will
    still be run.

.. note::

   If you can afford to add one extra 3rd party dependency, consider using the
   `~pyrseus.ctx.plugins.loky` plugin instead of this one. Here are a few of its
   advantages:

   - Its `~.loky.backend.context.cpu_count` function for determining the default
     ``max_workers`` is more sophisticated than ours, e.g. it respects not only
     the CPU affinity mask, but also cgroups constraints.

   - Even though the |loky|_ project has already led to significant improvements
     to the built-in `multiprocessing` library, its executors still handle a few
     worker crash situations better than the built-in
     `~concurrent.futures.ProcessPoolExecutor` that this plugin uses (as of
     Python 3.10).

   - It supports reusable worker pools that are especially convenient to use in
     interactive code.

   - It supports the `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode for faster
     teardowns when there are uncaught exceptions in the main program.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""


import importlib
from functools import cached_property

from pyrseus.ctx.plugins.process import EntryPoint as _BaseEntryPoint


class EntryPoint(_BaseEntryPoint):
    @cached_property
    def is_available(self):
        return all(
            importlib.util.find_spec(mod_name) is not None
            for mod_name in ("cloudpickle", "pyrseus.executors.cpprocess")
        )

    @cached_property
    def ExecutorCls(self):
        from pyrseus.executors.cpprocess import CpProcessPoolExecutor

        return CpProcessPoolExecutor


ENTRY_POINT = EntryPoint()



"""
Provides a simple serial executor that captures exceptions in the standard way.

This is useful:

- for running executor-based code serially without having to rewrite all of the
  control flow, and
- making it easy to use a debugger to trace into task functions.
"""

from concurrent.futures import Executor, Future
from typing import Callable, TypeVar

__all__ = ["InlineExecutor"]

# Represents the generic return type of a submitted callable.
Ret = TypeVar("Ret")


class InlineExecutor(Executor):
    def __init__(self):
        """
        Creates an `~concurrent.futures.Executor` that evaluates the task
        immediately upon submission, trapping exceptions like normal executors
        do.

        As with the only built-in within-process executor,
        `~concurrent.futures.ThreadPoolExecutor`, arbitrary callables can be
        used, including non-picklable ones like lambdas:

            >>> import os
            >>> with InlineExecutor() as exe:
            ...     non_picklable = lambda: os.getpid()   # can't be pickled
            ...     fut = exe.submit(non_picklable)       # but works anyway
            ...     worker_pid_is_my_pid = fut.result()   # no pickling error
            ...     assert os.getpid() == worker_pid_is_my_pid

        See :doc:`../plugins` for a list of related executors.
        """
        self._closing = False

    def submit(self, fcn: Callable[..., Ret], /, *args, **kwargs) -> Future[Ret]:
        """
        Immediately evaluates ``fcn(*args, **kwargs)`` and embeds the result in
        a `~concurrent.futures.Future`. As with standard executors, any
        exceptions that occur are caught and recorded in that returned
        `~concurrent.futures.Future`.
        """
        # This implementation is the same as NoCatchExecutor.submit, except this
        # one captures all exceptions in the returned
        # `~concurrent.futures.Future` object, like the standard
        # `~concurrent.futures` executors do.

        # To be consistent with `~concurrent.futures.ProcessPoolExecutor`,
        # disallow new submissions once shutdown has started.
        if self._closing:
            raise RuntimeError(
                "Submissions are not allowed to an executor that is shutting down."
            )

        fut: Future[Ret] = Future()
        try:
            # Eagerly execute the function. This makes it easy for users to use
            # `pdb.set_trace` and similar troubleshooting techniques to debug
            # `fcn`.
            result = fcn(*args, **kwargs)

        except BaseException as exc:
            fut.set_exception(exc)
        else:
            fut.set_result(result)

        return fut

    def shutdown(self, *args, **kwargs):
        """
        As a minor improvement on the base class' method, this override
        disallows submissions after a shutdown has been started. This can assist
        with finding bugs in user code.
        """
        self._closing = True
        super().shutdown(*args, **kwargs)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # OnError Modes
#
# This notebook discusses how to choose between `pyrseus.ctx.mgr.ExecutorCtx`'s three `OnError` modes for your application.
#
# Its `on_error` parameter controls what `ExecutorCtx.__exit__`
# does with its worker pool and any incomplete futures if an
# "uncaught exception" is flowing through it.

# %% [markdown]
# ## Setup
#
# Before running this notebook, make sure everything it depends on is installed:
#
#     # Modify the docs/requirements.txt path if you're running this
#     # command from anything except the repository's root directory.
#     python -m pip install -r docs/requirements.txt
#
# These are some imports that we'll use throughout this notebook:

# %% tags=["pyflyby-cell"]
import sys

sys.path.append("../../../src")  # assume we're running it from a Pyrseus source clone
import os
import time
from concurrent.futures import ThreadPoolExecutor

from pyrseus.ctx.mgr import ExecutorCtx

# %% [markdown]
# ## "Uncaught Exceptions"
#
# When we say "uncaught exception", we mean an exception that's raised in the
# with-statement and isn't caught and suppressed inside that with-statement. To
# be clear, we're not talking about exceptions inside tasks, unless those
# exceptions are rethrown, e.g. by calling `fut.result()`.
#
# Consider the following function that raises an exception:


# %%
def raises():
    raise RuntimeError("I am an exception!")


# %% [markdown]
# Here is one common source of "uncaught exceptions." The initial exception that
# `raises` raises is caught by the future, but then when we call `fut.result()`,
# we trigger a new exception which we don't catch inside `exe`'s with-statement.

# %%
# NOTE: Almost all of our test cells contain an extra try-except-else construct
# like this one. This is because we're demonstrating what happens with uncaught
# exceptions, but we want to avoid cluttering the output cells with all the
# tracebacks.
try:
    with ThreadPoolExecutor(1) as exe:
        # This statement will cause an exception to be raised in a worker
        # thread. But it'll be caught and bound to the future. We are *not*
        # talking about these cases.
        fut = exe.submit(raises)
        # That said, a common source of uncaught exceptions in an executor's
        # context are tasks' rethrown exceptions when accessing their result()
        # method. We *are* talking about these cases.
        fut.result()

except RuntimeError as ex:
    if not str(ex) == "I am an exception!":
        raise
else:
    raise AssertionError("This example failed.")

# %% [markdown]
#
# Another way to trigger an "uncaught exception" in an executor's
# context is to have some non-future code that does so.

# %%
try:
    with ThreadPoolExecutor(1) as exe:
        # Start a silly task
        fut = exe.submit(raises)
        # Directly trigger an "uncaught exception".
        raise RuntimeError("I am also an exception!")
except RuntimeError as ex:
    assert str(ex) == "I am also an exception!"
else:
    raise AssertionError("This example failed.")

# %% [markdown]
#
# Yet another way is to trigger a `KeyboardInterrupt` while an executor is
# running by pressing the "Interrupt the kernel" button in Jupyter. We don't
# show that here because this is meant to be viewable as a static document.
#
# Next, we'll discuss the three options that Pyrseus supports for dealing
# with these kinds of uncaught exceptions.

# %% [markdown]
# ## Wait: Fail-after Uncaught Exceptions
#
# All of the concurrent `ExecutorCtx` plugins support an  `on_error="wait"`
# mode.
#
# In this mode, the context manager will try to complete every submitted future
# before exiting, even if an uncaught exception is flowing through its
# `__exit__` method. This is useful for fire-and-forget jobs, and for jobs where
# you'll be inspecting each futures' final state after leaving the
# with-statement.
#
# For most underlying executors, this is their default behavior. For example,
# here's implicit `on_error="wait"` behavior in action with the built-in
# `ThreadPoolExecutor` class:

# %%
try:
    with ThreadPoolExecutor(1) as exe:
        # warm up the worker
        assert exe.submit(os.getpid).result() == os.getpid()
        t0 = time.time()
        # submit a slow tasks
        f0 = exe.submit(time.sleep, 0.2)
        # submit a bad task
        f1 = exe.submit(raises)
        # submit another slow tasks
        f2 = exe.submit(time.sleep, 0.2)
        # raise an exception
        raise RuntimeError("uncaught")

except Exception as ex:
    t1 = time.time()
    assert str(ex) == "uncaught", ex  # not "I am en exception!"
    assert 0.4 <= t1 - t0 <= 1.0, t1 - t0  # still ran all futures
    assert f0.exception(0) is None  # f0 is fully done
    assert "I am an exception!" in str(f1.exception(0))  # f1 is fully done
    assert f2.exception(0) is None  # f2 is fully done
else:
    raise AssertionError("This example failed.")

# %% [markdown]
#
# Here's the equivalent code using `ExecutorCtx`. There's nothing surprising
# here.

# %%
try:
    with ExecutorCtx("thread", 1, on_error="wait") as exe:  # <---- THE ONLY CHANGE
        assert exe.submit(os.getpid).result() == os.getpid()
        t0 = time.time()
        f0 = exe.submit(time.sleep, 0.2)
        f1 = exe.submit(raises)
        f2 = exe.submit(time.sleep, 0.2)
        raise RuntimeError("uncaught")
except Exception as ex:
    t1 = time.time()
    if not str(ex) == "uncaught":
        raise
else:
    raise AssertionError("This example failed.")
assert 0.4 <= t1 - t0 <= 1.0, t1 - t0
assert f0.exception(0) is None
assert "I am an exception!" in str(f1.exception(0))
assert f2.exception(0) is None

# %% [markdown]
# ## Cancel: Fail-faster for Uncaught Exceptions
#
# All of the concurrent `ExecutorCtx` plugins except for `ipyparallel` also
# support an `on_error="cancel"` mode.
#
# In this mode, the context manager will try to cancel all of its pending
# futures to provide fail-fast behavior, with two important caveats:
# - Futures whose tasks are already running cannot be cancelled.
# - Many underlying executors will pre-queue some of their futures to reduce
#   dispatcher latency. This makes those pre-queued futures uncancellable too,
#   even though they aren't actually running yet. For example,
#   `ProcessPoolExecutor` pre-queues a single task by default. `loky`'s
#   executors pre-queue many tasks (approximately 2x the number of CPU cores,
#   for various definitions of "cores").
#
# We first demonstrate the cancel behavior without `ExecutorCtx`. Unfortunately,
# it requires some mildly tricky extra boilerplate code. Furthermore, this first
# approach only works with executors that have a `shutdown` method that takes a
# `cancel_futures` argument.

# %%
try:
    with ThreadPoolExecutor(1) as exe:
        # To emulate ExecutorCtx's on_error="cancel" behavior, users can add
        # this extra try-except boilerplate to every with-statement they use
        # with an executor. It works, but it's verbose.
        try:
            assert exe.submit(os.getpid).result() == os.getpid()
            t0 = time.time()
            f0 = exe.submit(time.sleep, 0.2)
            f1 = exe.submit(raises)
            f2 = exe.submit(time.sleep, 0.2)
            # For this test, let's make sure the first task is running so we can
            # see it not get cancelled.
            time.sleep(0.1)
            raise RuntimeError("uncaught")
        except Exception:
            # This approach only works for executors that support a
            # cancel_futures argument to shutdown.
            exe.shutdown(wait=True, cancel_futures=True)
            # Don't forget to reraise the uncaught exception!
            raise

except Exception as ex:
    t1 = time.time()
    assert str(ex) == "uncaught", ex  # since we reraised
    # With ThreadPoolExecutor, there's no pre-queuing, so anything that wasn't
    # actually running at shutdown time is really cancelled. For most executors
    # though, this isn't true: at least one more, and potentially tens to
    # hundreds more futures will be uncancellable and will still start and
    # finish running before the shutdown is done.
    assert 0.2 <= t1 - t0 < 0.4, t1 - t0  # f1 ran but f2 was skipped
    assert f0.exception(0) is None  # f1 ran
    assert f1.cancelled()  # f1 was successfully cancelled
    assert f2.cancelled()  # f2 was successfully cancelled
else:
    raise AssertionError("This example failed.")

# %% [markdown]
#
# Here is a more general way to write the boilerplate when not using
# `ExecutorCtx`. This works with executors whose `shutdown` methods lack a
# `cancel_futures` argument. Note that like all cancel-based approaches, this
# doesn't sidestep the pre-queuing problem.

# %%
try:
    with ThreadPoolExecutor(1) as exe:
        # This try-except construct is a more general way of emulating
        # on_error="cancel". It can be more intrusive, but it doesn't require
        # exe.shutdown to take a cancel_futures argument.
        try:
            assert exe.submit(os.getpid).result() == os.getpid()
            t0 = time.time()
            # With this approach, we need to track all of our futures.
            futs = []
            futs.append(exe.submit(time.sleep, 0.2))
            futs.append(exe.submit(raises))
            futs.append(exe.submit(time.sleep, 0.2))
            time.sleep(0.1)
            raise RuntimeError("uncaught")
        except Exception:
            # This approach works for all executors, assuming that
            # at least some futures are left in a pending state.
            num_cancelled = 0
            for fut in futs:
                num_cancelled += fut.cancel()
            # Don't forget to reraise the uncaught exception!
            raise

except Exception as ex:
    t1 = time.time()
    if not str(ex) == "uncaught":  # since we reraised
        raise
else:
    raise AssertionError("This example failed.")
# See the previous code cell about ThreadPoolExecutor's behavior (tested here)
# vs. how most other executors pre-queue tasks.
assert num_cancelled == 2
assert 0.2 <= t1 - t0 < 0.4, t1 - t0
assert futs[0].exception(0) is None  # f1 ran
assert futs[1].cancelled()  # f1 was successfully cancelled
assert futs[2].cancelled()  # f2 was successfully cancelled

# %% [markdown]
#
# And here's how to get cancel behavior with `ExecutorCtx`, for plugins that
# support it: we just pass `on_error="cancel"` to the constructor. `ExecutorCtx`
# will internally use one of the two above approaches.

# %%
try:
    with ExecutorCtx("thread", 1, on_error="cancel") as exe:
        # No extra try-except needed here. ExecutorCtx does the work for us.
        assert exe.submit(os.getpid).result() == os.getpid()
        t0 = time.time()
        f0 = exe.submit(time.sleep, 0.2)
        f1 = exe.submit(raises)
        f2 = exe.submit(time.sleep, 0.2)
        time.sleep(0.1)
        raise RuntimeError("uncaught")

except Exception as ex:
    t1 = time.time()
    if not str(ex) == "uncaught":
        raise
else:
    raise AssertionError("This example failed.")
assert 0.2 <= t1 - t0 < 0.4, t1 - t0
assert f0.exception(0) is None
assert f1.cancelled()
assert f2.cancelled()

# %% [markdown]
# ## Kill: Fail-fast for Uncaught Exceptions
#
# Currently, only the `loky` plugin for `ExecutorCtx` supports an
# `on_error="kill"` mode.
#
# In this mode, the context manager will kill its workers' processes so that
# users don't have to wait for existing (and pre-queued) tasks to finish running
# before exiting the context. This makes it faster than cancel mode, but this
# may bypass cleanup routines like releasing locks inside the user's task code,
# within-task context managers, atexit handlers in the workers, etc.
#
# Here's a quick example showing how `loky`'s pre-queuing can make the
# previously-discussed cancel mode slow. If you want to make the slowness
# clearer, increase the `time.sleep` argument.

# %%
try:
    with ExecutorCtx("loky", 1, on_error="cancel") as exe:  # <---- CANCEL IS SLOW
        assert exe.submit(os.getpid).result() != os.getpid()
        t0 = time.time()
        # We need to submit a lot of tasks for this demo. Depending on the
        # number of cores available on the current box and whether we're
        # using loky's reusable vs. non-reusable mode, the pre-queue depth
        # can be hundreds of tasks long.
        futs = [exe.submit(time.sleep, 0.2)]
        for _ in range(1000):
            futs.append(exe.submit(os.getpid))
        time.sleep(0.1)
        raise RuntimeError("uncaught")

except Exception as ex:
    t1 = time.time()
    if str(ex) != "uncaught":
        raise
else:
    raise AssertionError("This example failed.")

# The wall time would be longer if the extra pre-queued tasks weren't trivial.
assert 0.2 <= t1 - t0 < 0.4, t1 - t0
assert futs[0].exception(0) is None
# We submitted enough that we'll get a mix between successful and cancelled
# futures.
num_success = 0
num_cancelled = 0
for fut in futs:
    assert fut.done()
    if fut.cancelled():
        num_cancelled += 1
    else:
        assert fut.exception(0) is None
        num_success += 1
# In loky's current implementation, there will always be at least 2 pre-queued
# tasks. Often it's tens or hundreds.
assert num_success >= 3
assert num_cancelled > 0

# %% [markdown]
#
# And here we see that its kill mode is fast. We didn't even wait for the first
# task to be completed to exit the with-statement. If you want to make the
# quickness clearer, increase the `time.sleep` argument and notice that the cell
# still runs just as fast.

# %%
try:
    with ExecutorCtx("loky", 1, on_error="kill") as exe:  # <----- KILL IS FAST
        assert exe.submit(os.getpid).result() != os.getpid()
        t0 = time.time()
        futs = [exe.submit(time.sleep, 0.2)]
        for _ in range(1000):
            futs.append(exe.submit(os.getpid))
        time.sleep(0.1)
        raise RuntimeError("uncaught")

except Exception as ex:
    t1 = time.time()
    if str(ex) != "uncaught":
        raise
else:
    raise AssertionError("This example failed.")

assert 0.1 <= t1 - t0 < 0.2, t1 - t0  # killed even before the first task is done
num_success = 0
num_cancelled = 0
num_exceptions = 0
for fut in futs:
    assert fut.done()
    if fut.cancelled():
        num_cancelled += 1
    elif fut.exception(0) is None:
        num_success += 1
    else:
        num_exceptions += 1
assert num_success == 0  # killed before even the first task is done
assert num_cancelled == 0  # killed, not cancelled
assert num_exceptions == len(futs)  # all killed



"""
Tests that basic loky-specific features are supported, if that library is
installed. Here we mostly test that the ``reusable`` feature functions as we
expect.
"""

import os
import platform
import re
import signal
import time
from contextlib import contextmanager
from functools import cache, partial
from multiprocessing import Manager

import psutil
import pytest

from pyrseus.core.sys import SignalHandlerCtx
from pyrseus.ctx.mgr import ExecutorCtx
from pyrseus.ctx.registry import is_plugin_available

# Magic global variable that skips the whole test module if the plugin isn't
# installed.
#
# https://docs.pytest.org/en/latest/example/markers.html#scoped-marking
pytestmark = pytest.mark.skipif(
    not is_plugin_available("loky"), reason="loky is not available"
)


@cache
def get_worker_id():
    proc = psutil.Process()
    return (proc.pid, proc.create_time())


def get_inner_func():
    def inner_func():
        return 123

    return inner_func


def worker_exists(worker_id):
    pid, create_time = worker_id
    try:
        proc = psutil.Process(pid)
    except Exception:
        return False
    else:
        return create_time == proc.create_time()


def get_abbrev_state(fut):
    """
    Summarizes the state of a Future.
    """
    run = fut.running()
    cancelled = fut.cancelled()
    done = fut.done()
    return {
        # We only enumerate the combinations we actually expect to see in these
        # tests.
        (False, False, False): "pending",
        (True, False, False): "running",
        (False, True, False): "cancel_pending",
        (False, True, True): "cancelled",
        (False, False, True): "success",
    }[run, cancelled, done]


def count_states(futs, sparse=False):
    counts = {
        "pending": 0,
        "running": 0,
        "cancel_pending": 0,
        "cancelled": 0,
        "success": 0,
    }
    for fut in futs:
        state = get_abbrev_state(fut)
        counts[state] += 1
    if sparse:
        counts = {k: v for k, v in counts.items() if v != 0}
    return counts


@pytest.mark.parametrize("reuse", (Ellipsis, False, True, "auto"))
@pytest.mark.parametrize(
    "reusable,expected",
    (
        (Ellipsis, "loky.process_executor.ProcessPoolExecutor"),
        (False, "loky.process_executor.ProcessPoolExecutor"),
        (True, "loky.*ReusablePoolExecutor"),
    ),
)
def test_reusable_vs_reuse_types(reusable, reuse, expected):
    """
    Verifies that the executor type we get back after entering ExecutorCtx is
    exactly what we expect with all combinations of the ``reusable`` and
    ``reuse`` parameters.
    """
    kwargs = {}
    if reusable is not Ellipsis:
        kwargs["reusable"] = reusable
    if reuse is not Ellipsis:
        kwargs["reuse"] = reuse
    with ExecutorCtx("loky", **kwargs) as exe:
        assert exe.__class__.__name__ == "OnExitProxy"
        underlying_cls = exe._exe.__class__
        fqn = f"{underlying_cls.__module__}.{underlying_cls.__name__}"
        assert re.match(expected, fqn), (reusable, reuse, expected, fqn)


def test_reusable_true():
    """
    Verifies that the reusable executors actually get reused.
    """
    with ExecutorCtx("loky", 1, reusable=True) as exe0:
        worker0 = exe0.submit(get_worker_id).result()
        assert get_worker_id() != worker0
    assert worker_exists(worker0)

    with ExecutorCtx("loky", 1, reusable=True) as exe1:
        assert exe1.submit(get_worker_id).result() == worker0
    assert worker_exists(worker0)

    exe1.shutdown(kill_workers=True)  # be a good test and clean up
    if platform.system() == "Windows":
        time.sleep(0.2)  # Windows lacks true SIGTERM, so give it a moment.
    assert not worker_exists(worker0)


def test_reuse_false():
    """
    Verifies that ``reuse=False`` actually causes a new reusable executor to be
    created.
    """
    with ExecutorCtx("loky", 1, reusable=True) as exe0:
        worker0 = exe0.submit(get_worker_id).result()
        assert get_worker_id() != worker0
        assert worker_exists(worker0)
    assert worker_exists(worker0)

    with ExecutorCtx("loky", 1, reusable=True, reuse=False) as exe1:
        assert not worker_exists(worker0)
        worker1 = exe1.submit(get_worker_id).result()
        assert get_worker_id() != worker1
        assert worker0 != worker1
    assert not worker_exists(worker0)
    assert worker_exists(worker1)

    with ExecutorCtx("loky", 1, reusable=True) as exe:
        assert exe.submit(get_worker_id).result() == worker1
    assert not worker_exists(worker0)
    assert worker_exists(worker1)

    exe1.shutdown(kill_workers=True)  # be a good test and clean up
    assert not worker_exists(worker0)
    assert not worker_exists(worker1)


def test_reuse_and_kill_workers():
    """
    Verifies that the ``reuse`` and ``kill_workers`` parameters interact the way
    we expect them to.
    """
    with ExecutorCtx("loky", 1, reusable=True) as exe0:
        worker0 = exe0.submit(get_worker_id).result()
        assert get_worker_id() != worker0
    assert worker_exists(worker0)

    # kill_workers is ignored when reuse=True
    with ExecutorCtx("loky", 1, reusable=True, reuse=True, kill_workers=True) as exe1:
        assert exe1.submit(get_worker_id).result() == worker0
    assert worker_exists(worker0)

    # but it does bounce the workers when reuse=False
    with ExecutorCtx("loky", 1, reusable=True, reuse=False, kill_workers=True) as exe2:
        worker2 = exe2.submit(get_worker_id).result()
        assert get_worker_id() != worker2
        assert worker0 != worker2
    assert not worker_exists(worker0)
    assert worker_exists(worker2)

    with ExecutorCtx("loky", 1, reusable=True) as exe3:
        assert exe3.submit(get_worker_id).result() == worker2
    assert not worker_exists(worker0)
    assert worker_exists(worker2)

    exe3.shutdown(kill_workers=True)  # be a good test and clean up
    if platform.system() == "Windows":
        time.sleep(0.2)  # Windows lacks true SIGTERM, so give it a moment.
    assert not worker_exists(worker0)
    assert not worker_exists(worker2)


def test_always_picklable_func():
    """
    Test that verifies that if our other picklability tests fail, it's due to a
    general problem, not a pickling-specific problem.
    """
    # If this crashes, everything will.
    with ExecutorCtx("loky", 1) as exe:
        exe.submit(os.getpid).result()


def test_import_env_transferred():
    """
    Verifies that our `sys.path` gets propagated to the workers with the
    ``"loky"`` plugin. If this fails, then the rest of the pickling tests may
    be failing for a more general reason that the specific picklability testing
    they're trying to do.
    """
    # This test module sits outside the normal import hierarchy, so this will
    # crash unless the remote workers have replicated the import environment of
    # the main process.
    with ExecutorCtx("loky", 1) as exe:
        exe.submit(get_worker_id).result()


def test_inner_func_pickling():
    """
    Verifies that non-global functions get pickled correctly. These have the
    same type as global functions but are not directly importable.
    """
    # If this fails, it means that the pickler didn't figure out that the
    # function isn't at the global scope of its module.
    with ExecutorCtx("loky", 1) as exe:
        func = get_inner_func()
        assert exe.submit(func).result() == 123


def test_lambda_pickling():
    """
    Verifies that lambda functions get pickled correctly. These have a different
    type than global or inner functions, so it's generally easier to get these
    working properly than global functions. E.g. these can sometimes work even
    if `sys.path` has not been propagated to the workers.
    """
    # If this fails, it means the pickler can't handle lambdas properly.
    with ExecutorCtx("loky", 1) as exe:
        assert exe.submit(lambda: 123).result() == 123
        assert exe.submit(lambda: get_inner_func()()).result() == 123
        func = get_inner_func()
        assert exe.submit(lambda: func()).result() == 123
        assert exe.submit(lambda func=func: func()).result() == 123


def test_double_picklability():
    """
    Verifies that a smart pickler like |cloudpickle|_ is being used not just for
    sending tasks, but also for getting their results.
    """
    # Here we verify that advanced pickling is available in both directions. We
    # send a lambda and get back a lambda.
    with ExecutorCtx("loky", 1) as exe:
        inner = lambda: 123  # NOQA
        outer = lambda: inner  # NOQA
        fut = exe.submit(outer)
        inner_ret = fut.result()
        ret = inner_ret()
        assert ret == 123


@contextmanager
def setup_reusable_test(max_workers):
    """
    Helper for tests that mix a reusable executor, a managed proxy of that same
    executor, and the ability to create shared barriers for testing.
    """
    from loky import get_reusable_executor

    kw = {"max_workers": max_workers, "timeout": 60}
    get_reusable = partial(get_reusable_executor, **kw)
    get_ctx = partial(ExecutorCtx, "loky", reusable=True, **kw)

    exe = None
    try:
        with Manager() as sync_mgr:
            yield get_reusable, get_ctx, sync_mgr.Barrier
    finally:
        if exe is not None:
            exe.shutdown(wait=True, kill_workers=True)


def test_reusable_can_exit_ctx_while_direct_fut_blocks():
    """
    Test for using a reusable executor both directly and via the context
    manager: verifies that we can exit the context while a direct fut is still
    running.
    """
    # Future work: troubleshoot this. Guess: this might be related to the
    # problem in test_ipyparallel_plugin's test_simple_args test on macOS.
    if platform.system() == "Darwin":
        pytest.skip("This test times out on macOS.")

    barrier_sets = []

    def get_next_waiter(*barriers):
        """
        Appends the supplied `barriers` as a tuple to `barrier_sets` and returns
        a lambda that will wait on that just-added set.
        """
        barrier_sets.append(barriers)
        return lambda: len([b.wait() for b in barriers])

    def get_num_waiting(num_waiters_to_ignore=0):
        """
        The number of barrier sets that have a least one waiting party, after
        taking into account ``num_waiters_to_ignore``.
        """
        num_waiting = 0
        for barriers in barrier_sets:
            if any(barrier.n_waiting > num_waiters_to_ignore for barrier in barriers):
                num_waiting += 1
        return num_waiting

    with setup_reusable_test(2) as (get_reusable, get_ctx, Barrier):
        # Request 2 workers.
        direct = get_reusable()

        # Hack to make the queue be small enough that all of our submitted tasks
        # don't get immediately slurped up and made uncancellable.
        super(type(direct), direct)._setup_queues(None, None, queue_size=1)

        # Submit a direct task that'll block on two barriers. The first will be
        # shared with the first ctx task to help us verify that Loky uses FIFO
        # scheduling. The second is used to control when this particular task is
        # allowed to finish.
        futs = [direct.submit(get_next_waiter(Barrier(3), Barrier(2)))]

        # Now create a context manager that reuses the same executor.
        with get_ctx() as ctx:
            try:
                # Succeeds because of the way we setup get_reusable and get_ctx'
                # arguments.
                assert ctx._exe is direct

                # Append one task that'll jointly wait with the direct task.
                futs.append(ctx.submit(get_next_waiter(barrier_sets[0][0], Barrier(2))))

                # Append a few more that should all be stuck behind the first
                # two. Make enough of them that we can successfully cancel some
                # later, despite Loky's aggressive latency-optimized queueing
                # system.
                for i in range(2, 100):
                    futs.append(ctx.submit(get_next_waiter(Barrier(2))))
                assert len(barrier_sets) == 100
                assert len(futs) == 100

                # Wait for the first two to be running. Note that we can't trust
                # the abbrev_state too much: Loky puts extra tasks in the
                # running state to reduce latency, even though they're not
                # actually running yet.
                assert barrier_sets[0][0] is barrier_sets[1][0]
                barrier_sets[0][0].wait(5)
                time.sleep(0.01)  # give both workers a moment to hit the next barrier
                assert get_abbrev_state(futs[0]) == "running"
                assert get_abbrev_state(futs[1]) == "running"
                assert barrier_sets[0][0].n_waiting == 0  # passed shared barrier
                assert barrier_sets[0][1].n_waiting == 1  # waiting on our signal
                assert barrier_sets[1][1].n_waiting == 1  # waiting on our signal
                assert get_num_waiting() == 2  # only the first 2 tasks waiting

                # Now let's try to cancel all of the actually-pending tasks so
                # that we don't deadlock the exit. Note that we *don't* wait on
                # their barriers. We want to verify that they get cancelled
                # before being scheduled.
                num_cancelled = 0
                for i in range(2, 100):
                    num_cancelled += futs[i].cancel()
                assert num_cancelled == 100 - 3  # 2 in progress, 1 pre-queued

                # Let fut 1 (only) return. We have to give a small timeout here
                # because we're doing sequential waits in the workers, not set
                # waits.
                barrier_sets[1][1].wait(1)
                assert futs[1].result() == 2  # task had 2 barriers

                # We also need to let the one pre-queued task also proceed since
                # we couldn't cancel it.
                barrier_sets[2][0].wait(5)
                assert futs[2].result() == 1  # task had 1 barrier

                # Exit the context.

            except BaseException:
                # If something in this test fails, immediately shut down
                # everything instead of being stuck in a deadlock we may have
                # created with our barriers.
                direct.shutdown(wait=True, kill_workers=True)
                raise

        # Immediately after exiting the context, we should be able to check the
        # states. There should be no intermediate states due to the wait call in
        # the context's __exit__.
        assert get_abbrev_state(futs[0]) == "running"
        assert get_abbrev_state(futs[1]) == "success"
        assert get_abbrev_state(futs[2]) == "success"
        counts = count_states(futs, sparse=True)
        assert counts == {"running": 1, "success": 2, "cancelled": 97}

        # Let the direct task finish.
        barrier_sets[0][1].wait(1)
        for _ in range(10):
            if get_abbrev_state(futs[0]) == "success":
                break
            time.sleep(0.1)
        assert get_abbrev_state(futs[0]) == "success"
        counts = count_states(futs, sparse=True)
        assert counts == {"success": 3, "cancelled": 97}

        direct.shutdown(wait=True, kill_workers=True)


def test_reusable_futs_wait_on_exit():
    """
    This is a companion test to test_can_exit_ctx_while_direct_fut_blocks. That
    one verified that __exit__ doesn't block on futures associated with the
    reusable executor but not with the ExecutorCtx. This one tests that __exit__
    does block on futures created by the ExecutorCtx.
    """
    if platform.system() == "Windows":
        pytest.skip("This test needs SIGALRM, but it is unavailable on Windows.")

    with setup_reusable_test(2) as (get_reusable, get_ctx, Barrier):
        # Here we're setting up a barrier that lets us control when our test
        # task is allowed to finish. We also setup an alarm system that lets us
        # release the barrier asynchronously.
        barrier = Barrier(2)
        barrier_state = "created"  # for troubleshooting

        def release_barrier_on_sigalrm(signum, frame):
            if signum == signal.SIGALRM:
                nonlocal barrier_state
                barrier_state = "waiting"
                barrier.wait()
                barrier_state = "released"

        with SignalHandlerCtx(signal.SIGALRM, release_barrier_on_sigalrm):
            with get_ctx() as exe:
                # Warm up the system.
                assert None is exe.submit(lambda: None).result()

                # Estimate the round trip time.
                t0 = time.time()
                assert None is exe.submit(lambda: None).result()
                t1 = time.time()
                rtt_secs = t1 - t0

                # Submit a task that'll block till we release it. But don't wait
                # on
                fut = exe.submit(barrier.wait)

                # Setup the alarm for a little while in the future. Make it long
                # enough that it's unlikely that the task would take that long
                # if we used a 1-party instead of 2-party barrier.
                wait_secs = max(1, int(rtt_secs * 4))
                earliest_exit_time = time.time() + wait_secs
                assert barrier_state == "created"
                signal.alarm(wait_secs)

            # Verify that __exit__ blocked till our task was done running.
            actual_exit_time = time.time()
            assert actual_exit_time >= earliest_exit_time
            assert barrier_state == "released"
            assert get_abbrev_state(fut) == "success"


def test_reusable_fut_leaks():
    """
    Verifies that in reusable mode, saved futures are quickly removed from the
    set of ones we save for __exit__-time cleanup.

    This is a whitebox text because it's a lot easier to write it that way, and
    we're testing an internal optimization.
    """
    with setup_reusable_test(1) as (get_reusable, get_ctx, Barrier):
        with get_ctx() as exe:
            # Submit a bunch of tasks that each block tell we let them proceed.
            barriers = []
            futs = []
            for _ in range(10):
                barriers.append(Barrier(2))
                futs.append(exe.submit(barriers[-1].wait))
            # All of them are in the cleanup set.
            assert exe._futs == set(futs)
            # Release the futures one at a time and verify that they get removed
            # from the cleanup set before we get them back.
            for i in range(10):
                assert exe._futs == set(futs[i:])
                barriers[i].wait(5)
                futs[i].result()
            assert len(exe._futs) == 0



"""
Provides a wrapper around `~pyrseus.ctx.mgr.ExecutorCtx` to make it useful in
interactive use cases.
"""
import sys

from pyrseus.ctx.mgr import ExecutorCtx


class _CtxMgrToRAIIProxy:
    """
    Wrapper proxy for a context manager that turns it into an `RAII
    <https://en.cppreference.com/w/cpp/language/raii>`_-like class.

    Background:

     - Python context managers are designed to be used in ``with`` statements,
       typically deferring heavy work till their ``__enter__`` method is called.
       When control flow leaves the ``with`` statement, the context manager's
       ``__exit__`` method is called, resulting in eager and reliable cleanup.

     - In contrast, C++ RAII classes perform the heavy initialization in their
       constructors and all of the shutdown in their destructors. This works
       well in C++ because (a) C++ uses block scoping instead of function
       scoping of variables, (b) it maintains a strong distinction between
       stack-based objects and heap-based ones, and (c) it has strong guarantees
       about how exceptions, constructors, and destructors interact. This makes
       it the the correct way to ensure eager and reliable cleanup in C++.

    What this proxy class does is to trigger the ``__enter__`` method in its
    constructor, proxy all public member access to the object returned by
    ``__enter__``, and then call ``__exit__`` at finalization time.

    See `.get_executor` for when to use this vs. not use it.
    """

    def __init__(self, ctx: ExecutorCtx):
        self._ctx = ctx
        self._obj = ctx.__enter__()

    def __del__(self):
        del self._obj
        self._ctx.__exit__(*sys.exc_info())

    def __enter__(self):
        raise RuntimeError(
            "This proxy should not be used as a context manager. "
            "Directly use the original context manager instead."
        )

    def __exit__(self, *exc_info):
        raise RuntimeError(
            "This proxy should not be used as a context manager. "
            "Directly use the original context manager instead."
        )

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(f"Only public members are proxied, not: {name}")
        return getattr(self._obj, name)


def get_executor(*args, **kwargs):
    """
    FOR INTERACTIVE USE ONLY: This function is like `.ExecutorCtx`, but
     - this function pre-calls the executor after ``__enter__`` method and
     - attempts to perform its ``__exit__`` behavior in a ``__del__`` finalizer.

    This can be useful for interactive use cases where the worker pool should
    stay alive till the whole process exits.

    All of this function's arguments are passed to `.ExecutorCtx`.

    Here is an example of the same executor being used across multiple blocks of
    code. We explicitly delete the executor at the end to encourage eager
    cleanup of our worker pool.

        >>> exe = get_executor("thread", 1)
        >>> futs = [exe.submit(lambda x: x, x) for x in range(10)]
        >>> print(sum(fut.result() for fut in futs))
        45

        >>> futs = [exe.submit(lambda x: x*x, x) for x in range(10)]
        >>> print(sum(fut.result() for fut in futs))
        285

        >>> del exe  # try to eagerly cleanup

    .. note::

        If you have an interactive use case like an IPython session or a Jupyter
        notebook where:

        - you will be submitting parallel work in multiple different cells,
        - your worker pool startup time is non-trivial and/or you have complex
          parameters such as an initializer function, *and*
        - you don't mind the worker pool staying alive and consuming resources
          until the interactive process fully exits, then

        this function may be right for you. But if even one of those statements
        is false, consider using the more robust `.ExecutorCtx` instead in a
        proper ``with`` block.

        See the |nb_get_executor|_ notebook for additional guidance about the
        tradeoffs between `.ExecutorCtx` and `.get_executor`.

    """
    return _CtxMgrToRAIIProxy(ExecutorCtx(*args, **kwargs))



"""
Tests that basic cpmpi4py-specific features are supported, if that library is
installed.
"""

import os
from functools import cache

import psutil
import pytest

from pyrseus.ctx.mgr import ExecutorCtx
from pyrseus.ctx.registry import is_plugin_available

# Magic global variable that skips the whole test module if the plugin isn't
# installed.
#
# https://docs.pytest.org/en/latest/example/markers.html#scoped-marking
pytestmark = pytest.mark.skipif(
    not is_plugin_available("cpmpi4py"), reason="cpmpi4py is not available"
)


@cache
def get_worker_id():
    proc = psutil.Process()
    return (proc.pid, proc.create_time())


def get_inner_func():
    def inner_func():
        return 123

    return inner_func


def test_always_picklable_func():
    # If this crashes, everything will.
    with ExecutorCtx("cpmpi4py", 1) as exe:
        exe.submit(os.getpid).result()


@pytest.mark.slow
def test_import_env_transferred():
    # This test module sits outside the normal import hierarchy, so this will
    # crash unless the remote workers have replicated the import environment of
    # the main process.
    with ExecutorCtx("cpmpi4py", 1) as exe:
        exe.submit(get_worker_id).result()


@pytest.mark.slow
def test_inner_func_pickling():
    # If this fails, it means that the pickler didn't figure out that the
    # function isn't at the global scope of its module.
    func = get_inner_func()

    with ExecutorCtx("cpmpi4py", 1) as exe:
        assert exe.submit(lambda: 123).result() == 123
        assert exe.submit(lambda: get_inner_func()()).result() == 123
        func = get_inner_func()
        assert exe.submit(lambda: func()).result() == 123
        assert exe.submit(lambda func=func: func()).result() == 123


@pytest.mark.slow
def test_lambda_pickling():
    # If this fails, it means the pickler can't handle lambdas properly.
    with ExecutorCtx("cpmpi4py", 1) as exe:
        assert exe.submit(lambda: 123).result() == 123


def test_double_picklability():
    # Here we verify that advanced pickling is available in both directions. We
    # send a lambda and get back a lambda.
    with ExecutorCtx("cpmpi4py", 1) as exe:
        inner = lambda: 123  # NOQA
        outer = lambda: inner  # NOQA
        fut = exe.submit(outer)
        inner_ret = fut.result()
        ret = inner_ret()
        assert ret == 123



from unittest import TestCase

from pyrseus.executors.cpnocatch import CpNoCatchExecutor

Cls = CpNoCatchExecutor


def picklable():
    return 1


def make_lambda():
    return lambda: 2


def make_double_unpicklable():
    return lambda: lambda: 3


class TestPickling(TestCase):
    def test_picklable(self):
        with Cls() as exe:
            fut = exe.submit(picklable)
            ret = fut.result()
            assert ret == 1

    def test_lambda(self):
        with Cls() as exe:
            lambd = make_lambda()
            fut = exe.submit(lambd)
            ret = fut.result()
            assert ret == 2

    def test_return_lambda(self):
        with Cls() as exe:
            fut = exe.submit(make_lambda)
            ret_func = fut.result()
            ret = ret_func()
            assert ret == 2

    def test_double_unpicklable(self):
        with Cls() as exe:
            func = make_double_unpicklable()
            fut = exe.submit(func)
            ret_func = fut.result()
            ret = ret_func()
            assert ret == 3



"""
Helpers for writing `~concurrent.futures.Executor` classes and
`~pyrseus.ctx.mgr.ExecutorCtx` plugins that need to interact with `pickle`-based
serialization.

There are two main use cases:

- To change the serializer to a more powerful one like |cloudpickle|_. For
  examples, see the source code of:

  - `pyrseus.executors.cpprocess`: wrapper for
    `~concurrent.futures.ProcessPoolExecutor` that uses |cloudpickle|_ instead
    of `pickle` for task and result serialization.

  - `pyrseus.ctx.plugins.cpmpi4py`: wrapper for
    `~mpi4py.futures.MPIPoolExecutor` that uses |cloudpickle|_ instead of
    `pickle` for task and result serialization.

- To help users troubleshoot serialization problems. For examples, see the
  source code of:

  - `pyrseus.executors.pinline`: serializes tasks and results in the same thread
    and process as the ``submit`` call, using `pickle`.

  - `pyrseus.executors.cpinline`:  serializes tasks and results in the same
    thread and process as the ``submit`` call, using |cloudpickle|_.
"""

import inspect
import pickle
import sys
from functools import cache
from threading import Lock
from typing import Any, Callable, Dict, Tuple, TypeVar

Ret = TypeVar("Ret")
"""
Represents the generic return type of a submitted callable.
"""


class _HideMain:
    """
    Helper for `.try_pickle_round_trip` that temporarily replaces
    ``sys.modules['__main__']``, causing any attribute access to raise an
    exception. This can make troubleshooting pickling problems easier.
    """

    def __init__(self, hide_from):
        self._hide_from = hide_from

    def __getattr__(self, name):
        raise RuntimeError(
            f"Blocked attempt to read {name} from __main__ while pickling "
            f"{self._hide_from}."
        )


_hide_main_lock = Lock()
"""
Protects against concurrent calls to `.try_pickle_round_trip` from permanently
corrupting ``sys.modules``.
"""


def try_pickle_round_trip(
    obj, *, dumps=pickle._dumps, protocol=-1, loads=pickle._loads, hide_main=True
):
    """
    Attempts to pickle and unpickle ``obj`` using ``dumps`` and ``loads``, to
    help with testing and/or troubleshooting.

    Example
    -------

    Here's an example of an object that can be pickled and unpickled without
    trouble:

        >>> try_pickle_round_trip(42)
        42

    Here's one that acts like it's picklable and unpicklable:

        >>> # simulate defining a trivial function in __main__, as happens
        >>> # for scripts and notebooks.
        >>> exec(
        ...     'def _func_in_main_for_try_pickle_round_trip(): return 42',
        ...     sys.modules['__main__'].__dict__,
        ... )
        >>> # retrieve it
        >>> func = sys.modules['__main__']._func_in_main_for_try_pickle_round_trip

        >>> # technically, it's picklable and unpicklable, as long as the
        >>> # unpickling happens in the same process:
        >>> pickle.loads(pickle.dumps(func, -1))()
        42

    But if the unpickling happens in another process that doesn't replicate the
    original's ``__main__`` module, then unpickling will fail, often with a
    confusing error message.

    But with this function, we can simulate those unpickling problems, but at
    pickling time when they're easier to troubleshoot:

        >>> # Here it fails with a slightly better error message, and by
        >>> # default using the pure Python pickler that's easier to debug
        >>> # than the C one.
        >>> try_pickle_round_trip(func)
        Traceback (most recent call last):
        ...
          File ...pickle.py... in save_global
        ...
        RuntimeError: Blocked attempt to read
        _func_in_main_for_try_pickle_round_trip from __main__ while pickling
        <function _func_in_main_for_try_pickle_round_trip at ...>.

        >>> # Clean up
        >>> del sys.modules['__main__']._func_in_main_for_try_pickle_round_trip

    .. warning::

       When ``hide_main=True``, this function mutates global state by
       temporarily replacing ``sys.modules['__main__']``. This function is
       threadsafe with respect to itself, but if some other concurrent code
       requires access to the ``__main__`` module *via* ``sys.modules``, then
       they may see unexpected exceptions. Such situations are rare.

    :param obj: the object whose picklability is being tested

    :param protocol: pickle protocol number to use. This is passed to ``dumps``
        as a keyword argument.

    :param dumps: a function that serializes objects to a pickle bytestring.
        This defaults to the pure Python `pickle._dumps` function. It is slower
        than the C `pickle.dumps` function, but it's easier to troubleshoot with
        a Python debugger.

    :param loads: a function that loads objects from a pickle bytestring. This
        defaults to the pure Python `pickle._loads` function. It is slower than
        the C `pickle.loads` function, but it's easier to troubleshoot with a
        Python debugger.

    :param hide_main: whether to hide the ``__main__`` module during the pickle
        operation. This is done by temporarily replacing
        ``sys.modules['__main__']`` with a proxy object that raises an exception
        on any attribute access. This makes it much easier to detect and
        troubleshoot pickling problems that arise from objects that belong to a
        script or notebook instead of a Python module, without needing to
        perform the unpickling in a separate process where it's harder to debug.
    """
    if hide_main:
        with _hide_main_lock:
            old_main = sys.modules["__main__"]
            try:
                sys.modules["__main__"] = _HideMain(obj)
                pickled = dumps(obj, protocol=protocol)
                reconstructed = loads(pickled)
            finally:
                sys.modules["__main__"] = old_main
    else:
        pickled = dumps(obj, protocol=protocol)
        reconstructed = loads(pickled)
    return reconstructed


@cache
def get_round_trip_keywords():
    """
    Helper for creating plugins that wrap `try_pickle_round_trip` and/or
    `call_with_round_trip_pickling`.

    >>> sorted(get_round_trip_keywords())
    ['dumps', 'hide_main', 'loads', 'protocol']
    """
    return {
        p.name
        for p in inspect.signature(try_pickle_round_trip).parameters.values()
        if p.kind is inspect.Parameter.KEYWORD_ONLY
    }


def call_with_round_trip_pickling(func, args, kwargs, **round_trip_kwargs):
    """
    Calls ``func(*args, **kwargs)``, but (a) running those three things through
    `try_pickle_round_trip` first, and then (b) running the result through
    `try_pickle_round_trip` too.

    This function is designed for serial executors that help detect picklability
    problems in submitted tasks.

    :param func: function to call after round-trip pickling it

    :param args: positional arguments to pass to ``func`` after round trip
        pickling

    :param kwargs: keyword arguments to pass to ``func`` after round trip
        pickling

    :param round_trip_kwargs: passed to `try_pickle_round_trip` to override the
        round trip pickling settings

    :return: the return value of ``func(*args, **kwargs)``, after doing the two
        round trip pickling tests
    """
    r_func, r_args, r_kwargs = try_pickle_round_trip(
        (func, args, kwargs), **round_trip_kwargs
    )
    orig_ret = r_func(*r_args, **r_kwargs)
    ret = try_pickle_round_trip(orig_ret, **round_trip_kwargs)
    return ret


class OncePickledObject:
    """
    Creates a wrapper around an arbitrary Python object that forces it to be
    pickled with the chosen pickler, but a copy of the original unwrapped object
    is returned at unpickling time.

    This wrapper does not provide any proxying of attributes or methods. It is
    only intended to be used when users know the object will be pickled and
    unpickled exactly once before use.

    >>> opo123 = OncePickledObject(123, pickle.dumps, pickle.loads)
    >>> opo123
    <pyrseus.core.pickle.OncePickledObject object at ...>
    >>> pickle.loads(pickle.dumps(opo123, -1))
    123
    """

    def __init__(self, obj, dumps, loads):
        self._obj = obj
        self._dumps = dumps
        self._loads = loads

    def __reduce_ex__(self, protocol: int):
        # Pickle the object we're wrapping, using the chosen pickler, returning
        # a bytestring.
        pickled = self._dumps(self._obj, protocol=protocol)
        # Tell the pickler that's calling us that it can reconstruct the
        # original unwrapped object by calling the chosen unpickler on the
        # bytestring created from the original object we wrapped.
        return self._loads, (pickled,)


class CustomPickledClosure:
    def __init__(
        self,
        func: Callable,
        args: Tuple,
        kwargs: Dict,
        dumps: Callable[..., bytes],
        loads: Callable[[bytes], Any],
    ):
        """
        Wraps a function and its arguments into a nullary closure that forces a
        chosen pickler to be used for the closure and its return value.

        - The closed function will remain wrapped by this class when it
          undergoes round trip pickling.

        - The return value of the closed function will be wrapped in a
          `.OncePickledObject` that uses the same ``dumps`` and ``loads``
          functions as the wrapper.

        This is useful for wrapping functions that will be submitted to
        executors like `~concurrent.futures.ProcessPoolExecutor` with workers in
        other processes. This forces the chosen pickler to be used for
        serialization, instead of whatever default pickler the executor normally
        uses.

        :param func: the function whose pickling behavior we're overriding

        :param args: the positional arguments to ``func`` that we're closing
            over

        :param kwargs: the keyword arguments to ``func`` that we're closing over

        :param dumps: a `pickle.dumps`-like function that will be used for
            serializing the closure contents when this wrapper is pickled. This
            function must accept a ``protocol`` keyword argument.

        :param loads: a `pickle.loads`-like function that will be used for
            deserializing the closure contents when the pickled form of this
            wrapper is unpickled. To avoid bootstrapping problems, this function
            must itself be picklable by the built-in `pickle.dumps`.
        """
        self._func = func
        self._args = args
        self._kwargs = kwargs
        self._dumps = dumps
        self._loads = loads

    def __reduce_ex__(self, protocol: int):
        # First, we need to pickle up the unpickler. To avoid bootstrapping
        # problems if the unpickling happens in another process, this is done
        # with pickle.dumps itself.
        pickled_loads = pickle.dumps(self._loads, protocol=protocol)
        # Now pickle up the whole closure, using the requested serializer. For
        # efficiency, we don't re-serialize self._loads
        args = [self._func, self._args, self._kwargs, self._dumps]
        dumped_args = self._dumps(args, protocol=protocol)
        # This instructs the unpickler to use our special factory that undoes
        # the extra nested pickling.
        return type(self)._hydrate, (pickled_loads, dumped_args)

    @classmethod
    def _hydrate(cls, pickled_loads: bytes, dumped_args: bytes):
        loads = pickle.loads(pickled_loads)
        hydrated_args = loads(dumped_args)
        hydrated_args.append(loads)
        return cls(*hydrated_args)

    def __call__(self):
        # Call the underlying (reconstructed) function.
        raw_ret = self._func(*self._args, *self._kwargs)
        # Return a wrapper object that will produce a copy of the underlying
        # object when pickled and unpickled.
        return OncePickledObject(raw_ret, self._dumps, self._loads)



"""
The `pyrseus` package provides the `~pyrseus.ctx.mgr.ExecutorCtx` factory
class, its plugins, and various standalone serial `~concurrent.futures.Executor`
classes. Its goals are to make it easier to troubleshoot executor-based code by
making it easy switch between executor backends, including serial executors
provided by this package.
"""

# Re-export the most useful things.
from pyrseus.ctx.mgr import ExecutorCtx  # NOQA
from pyrseus.executors.cpinline import CpInlineExecutor  # NOQA
from pyrseus.executors.cpnocatch import CpNoCatchExecutor  # NOQA
from pyrseus.executors.cpprocess import CpProcessPoolExecutor  # NOQA
from pyrseus.executors.inline import InlineExecutor  # NOQA
from pyrseus.executors.nocatch import NoCatchExecutor  # NOQA
from pyrseus.executors.pinline import PInlineExecutor  # NOQA
from pyrseus.executors.pnocatch import PNoCatchExecutor  # NOQA

__version__ = "0.1.0"



"""
Provides a variant of `InlineExecutor` that tests the picklability of all
submitted tasks and their return values, using `pickle`.
"""

from functools import partial

from pyrseus.core.pickle import call_with_round_trip_pickling
from pyrseus.executors.inline import InlineExecutor


class PInlineExecutor(InlineExecutor):
    def __init__(self, **round_trip_kwargs):
        """
        Pickle-testing variant of `.InlineExecutor`, using the built-in `pickle`
        module.

        This variant pickles each submitted task and the task's results, using
        the built-in `pickle` module. This is primarily useful for
        troubleshooting pickling problems occurring in multi-process executors,
        by performing the pickling and unpickling locally.

        Here's an example showing a pickling failure cause by trying to pickle a
        lambda with an executor type that uses `pickle` instead of
        |cloudpickle|_.

        Consider this lambda.

            >>> import pickle
            >>> needs_cloudpickle = lambda: "this only works with cpnocatch, not pnocatch"

        It can't be pickled with `pickle`.

            >>> pickle.dumps(needs_cloudpickle, -1)
            Traceback (most recent call last):
            ...
            _pickle.PicklingError: Can't pickle ...

        Since this class tests pickling with `pickle`, it correctly fails, as it
        would in a multi-process executor that uses `pickle`.

            >>> with PInlineExecutor() as exe:
            ...     fut = exe.submit(needs_cloudpickle)
            >>> # As with InlineExecutor, the exception is only propagated out
            >>> # at result() time.
            >>> print(fut.result())
            Traceback (most recent call last):
            ...
            _pickle.PicklingError: Can't pickle ...

        Because the failure happens within the same process and thread as the
        submit call, it is easy to debug by tracing into it with a debugger.

        See :doc:`../plugins` for a list of related executors.

        :param round_trip_kwargs: keyword arguments passed to
            `call_with_round_trip_pickling` for each submitted task.
        """
        super().__init__()
        self._round_trip_kwargs = round_trip_kwargs

    def submit(self, fcn, /, *args, **kwargs):
        testing_closure = partial(
            call_with_round_trip_pickling,
            fcn,
            args,
            kwargs,
            **self._round_trip_kwargs,
        )
        return super().submit(testing_closure)
